/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/
/***************************************************************************
OctreeCHCcamera.h  -  description
-------------------
begin                : Fri Sep 27 2002
copyright            : (C) 2002 by Jon Anderson
email                : janders@users.sf.net

Enhancements 2003 - 2004 (C) The OGRE Team
***************************************************************************/

#ifndef OctreeCHCCAMERA_H
#define OctreeCHCCAMERA_H

#include "OctreeCHCPrerequisites.h"

#include "OgreCamera.h"
#include "OgreHardwareBufferManager.h"
#include "OgreSimpleRenderable.h"
#include "OgreMath.h"
#include "OgreAxisAlignedBox.h"
#include "OgreRoot.h"
#include "OgreViewport.h"
#include "OgreRenderSystem.h"


#include "RadarFrustum.h"
#include "OctreeCHCOcclusionVisibilityData.h"
/**
*@author Jon Anderson
*/

namespace OgreOctreeCHC
{

	class OctreeCHC;


	/** Specialized viewpoint from which an OctreeCHC can be rendered.
	@remarks
		This class contains several specializations of the Ogre::Camera class. It
		implements the getRenderOperation method in order to return displayable geometry
		for debugging purposes. It also implements a visibility function that is more granular
		than the default.
	*/

	class OctreeCHCCamera : public Ogre::Camera
	{

    public:

			class Cone
			{
			public:

				Ogre::Vector3 getAxis() const { return mAxis; }
				void setAxis(const Ogre::Vector3 &val) { mAxis = val; }
				Ogre::Vector3 getVertex() const { return mVertex; }
				void setVertex(const Ogre::Vector3 &val) { mVertex = val; }
				Ogre::Radian getConeAngle() const { return mConeAngle; }
				void setConeAngle(Ogre::Radian val) 
				{ 
					mConeAngle = val; 

					mCosSqr = Ogre::Math::Cos(val);
					mCosSqr = mCosSqr*mCosSqr;

					mSinSqr = Ogre::Math::Sin(val);
					mSinReciprocal = 1.0 / mSinSqr;
					mSinSqr = mSinSqr*mSinSqr;
				}

				bool intersects (const Ogre::Sphere &sphere) const 
				{
					Ogre::Vector3 U = mVertex - (sphere.getRadius()*mSinReciprocal)*mAxis;
					Ogre::Vector3 D (sphere.getCenter() - U);
					Ogre::Real dsqr = D.dotProduct(D);
					Ogre::Real e = mAxis.dotProduct(D);
					if ( e > 0 && e*e >= dsqr*mCosSqr )
					{
						D = sphere.getCenter() - mVertex;
						dsqr = D.dotProduct(D);
						e = - mAxis.dotProduct(D);
						if ( e > 0 && e*e >= dsqr*mSinSqr )
							return dsqr <= sphere.getRadius() * sphere.getRadius();
						else
							return true;
					}
					return false;
				}

			public:
				Ogre::Real mSinReciprocal;
				Ogre::Real mCosSqr;
				Ogre::Real mSinSqr;

				Ogre::Vector3 mAxis;
				Ogre::Vector3 mVertex;
				Ogre::Radian mConeAngle;
			};



            /** Visibility types */
            enum Visibility
            {
                NONE,
                PARTIAL,
                FULL
            };

            /* Standard Constructor */
            OctreeCHCCamera(const Ogre::String& name, Ogre::SceneManager* sm);
            /* Standard destructor */
            virtual ~OctreeCHCCamera(void);

			//-----------------------------------------------------------------------
			void updateCameraPrecalc();

			/** Returns the visibility of the Sphere
			*/
			//-----------------------------------------------------------------------
			Visibility getVisibility(const Ogre::Sphere& refSphere) const
			{
				const Ogre::Real radius (refSphere.getRadius());
				const Ogre::Vector3 &center (refSphere.getCenter());
				// calculate our distances to each of the planes
				const bool infinite_far_clip = (mFarDist == 0);
				for (unsigned int plane = 0; plane < 6; ++plane)
				{
					// Skip far plane if infinite view frustum
					if (infinite_far_clip && Ogre::FRUSTUM_PLANE_FAR == plane)
						continue;

					// This updates frustum planes and deals with cull frustum
					const Ogre::Real dist = getFrustumPlane(plane).getDistance(center);
					// if this distance is < -sphere.radius, we are outside
					if (dist < -radius)
						return NONE;

					// else if the distance is between +- radius, then we intersect
					if (Ogre::Math::Abs(dist) < radius)
						return PARTIAL;					
				}
				return FULL;
			}
			/** Returns the visibility of the OcclusionElement
			*/
			//-----------------------------------------------------------------------
			inline Visibility  getVisibility(const OcclusionElement &oElement) const
			{

				// box is not hierarchically culled as inside frustum box
				// so we need to check it.
				const Ogre::AxisAlignedBox &boxBounds = oElement.getBoundingBox ();

				// Null boxes always invisible
				if (boxBounds.isNull())
					return NONE;
				// Infinite boxes always visible
				if (boxBounds.isInfinite())
					return FULL;


				if (!boxBounds.contains(getRealPosition()))// if camera is in box, then box is visible
				{
					const Ogre::Sphere &sphereBounds = oElement.getBoundingSphere();

					// check if we are in the sphere of the frustum
					if (!(mFrustumSphere.intersects(sphereBounds)))
						return NONE;
					// check if we are in the cone of the frustum
					if (!(mFrustumCone.intersects(sphereBounds)))
						return NONE;

					// test the bounding sphere first
					{
						Visibility sphereTestResult = getVisibility(sphereBounds); 
						if (sphereTestResult != PARTIAL)
							return sphereTestResult;
					}
					return getVisibility (boxBounds);
				}
				return PARTIAL;
			}
			/** Returns the visibility of the box
			*/
			//Visibility getVisibility(const Ogre::AxisAlignedBox& bound) const;
			//-----------------------------------------------------------------------
			inline bool OctreeCHCCamera::isVisibleOnRadar(const OcclusionElement &oElement) const 
			{
				// box is not hierarchically culled as inside frustum box
				// so we need to check it.
				const Ogre::AxisAlignedBox &boxBounds = oElement.getBoundingBox ();

				// Null boxes always invisible
				if (boxBounds.isNull())
					return false;
				// Infinite boxes always visible
				if (boxBounds.isInfinite())
					return true;

				if (!boxBounds.contains(getRealPosition()))// if camera is in box, then box is visible
				{
					const Ogre::Sphere &sphereBounds = oElement.getBoundingSphere();

					// check if we are in the sphere of the frustum
					if (!(mFrustumSphere.intersects(sphereBounds)))
						return false;
					// check if we are in the cone of the frustum
					if (!(mFrustumCone.intersects(sphereBounds)))
						return false;

					if (mRadar.IsSphereIn (oElement.getBoundingSphere().getCenter(), 
										   oElement.getBoundingSphere().getRadius()))
					{
						if (mRadar.IsAABBIn (boxBounds.getCenter() - boxBounds.getHalfSize(), 
											 boxBounds.getCenter() + boxBounds.getHalfSize()))
							return true;
						else
							return false;
					}
					else
					{
						return false;
					}
				}
				return true;
			}
            /** Returns the visibility of the box
            */
            //Visibility getVisibility(const Ogre::AxisAlignedBox& bound) const;
			//-----------------------------------------------------------------------
			inline Visibility OctreeCHCCamera::getVisibility(const Ogre::AxisAlignedBox &bound) const 
			{
				// Null boxes always invisible
				if ( bound.isNull() )
					return NONE;
				// Infinite boxes always visible
				if (bound.isInfinite())
					return FULL;

			
				// Get centre of the box
				const Ogre::Vector3 &centre (bound.getCenter());
				// Get the half-size of the box
				const Ogre::Vector3 &halfSize (bound.getHalfSize());

				bool all_inside = true;

				const bool infinite_far_clip = (mFarDist == 0);
				for (unsigned int plane = 0; plane < 6; ++plane)
				{
					// Skip far plane if infinite view frustum
					if (infinite_far_clip && Ogre::FRUSTUM_PLANE_FAR == plane)
						continue;

					// This updates frustum planes and deals with cull frustum
					//const Ogre::Plane::Side side = mFrustumPlanes[plane].getSide(centre, halfSize);
					const Ogre::Plane::Side side = getFrustumPlane(plane).getSide(centre, halfSize);
					if (side == Ogre::Plane::NEGATIVE_SIDE) 
						return NONE;
					// We can't return now as the box could be later on the negative side of a plane.
					if (side == Ogre::Plane::BOTH_SIDE) 
						all_inside = false;
				}
				if ( all_inside )
					return FULL;
				else
					return PARTIAL;

			}


			//-----------------------------------------------------------------------
			inline Ogre::Vector3  getFarestPoint (const size_t i, const Ogre::AxisAlignedBox & box)
			{
				// should really pre-compute a bitfield of sign per Plane Frustum.
				return Ogre::Vector3
					(
						((mPlaneNormalSign[i] & 1) ? box.getMaximum().x : box.getMinimum().x),
						((mPlaneNormalSign[i] & 2) ? box.getMaximum().y : box.getMinimum().y),
						((mPlaneNormalSign[i] & 4) ? box.getMaximum().z : box.getMinimum().z)					
					);
			}
			//-----------------------------------------------------------------------
			inline Ogre::Vector3 getNearestPoint (const size_t i, const Ogre::AxisAlignedBox & box)
			{
				// should really pre-compute a bitfield of sign per Plane Frustum.
				return Ogre::Vector3
					(
						((!(mPlaneNormalSign[i] & 1)) ? box.getMaximum().x : box.getMinimum().x),
						((!(mPlaneNormalSign[i] & 2)) ? box.getMaximum().y : box.getMinimum().y),
						((!(mPlaneNormalSign[i] & 4)) ? box.getMaximum().z : box.getMinimum().z)					
					);
			}
			//-----------------------------------------------------------------------
			Visibility getHierarchicalVisibility(OcclusionElement &oElement,  size_t parentMask, size_t *childrenMask)
			{
				Visibility result = FULL;

				const Ogre::AxisAlignedBox &box = oElement.getBoundingBox();
				*childrenMask = 0;


				for (size_t i = 0, k = 1; i < 6; i++, k += k)
				{
					if (k & parentMask) 
					{ 
						const Ogre::Plane &sp = getFrustumPlane((unsigned short)i);
						if (sp.getSide(getFarestPoint(i, box)) == Ogre::Plane::NEGATIVE_SIDE) 
						{
							return NONE;
						}
					} 
				}

				for (size_t i = 0, k = 1; i < 6; i++, k += k)
				{
					if (k & parentMask) 
					{ 
						const Ogre::Plane &sp = getFrustumPlane((unsigned short)i);

						if (sp.getSide(getNearestPoint(i, box)) == Ogre::Plane::NEGATIVE_SIDE) 
						{
							*childrenMask |= k;
							result = PARTIAL;
						}
					} 
				}

				return result;				
			}
			//-----------------------------------------------------------------------
			Visibility getHierarchicalAndTemporalVisibility(OcclusionElement &oElement,  size_t parentMask, size_t *childrenMask)
			{
				Visibility result = FULL;

				const size_t startId = oElement.currVis->lastplaneCulledById;
				const Ogre::AxisAlignedBox &box = oElement.getBoundingBox();
				*childrenMask = 0;

				size_t k = 1 << startId;

				if (k & parentMask) 
				{
					const Ogre::Plane &sp = getFrustumPlane((unsigned short)startId);
					if (sp.getSide(getFarestPoint(startId, box)) == Ogre::Plane::NEGATIVE_SIDE) 
						return NONE;

					if (sp.getSide(getNearestPoint(startId, box)) == Ogre::Plane::NEGATIVE_SIDE) 
					{
						*childrenMask |= k;
						result = PARTIAL;
					}
				}

				for (size_t i = 0, k = 1; i < 6; i++, k += k)
				{
					if ((i != startId) && (k & parentMask)) 
					{ 
						const Ogre::Plane &sp = getFrustumPlane((unsigned short)i);
						if (sp.getSide(getFarestPoint(i, box)) == Ogre::Plane::NEGATIVE_SIDE) 
						{
							oElement.currVis->lastplaneCulledById = i;
							return NONE;
						}
					} 
				}	
				
				for (size_t i = 0, k = 1; i < 6; i++, k += k)
				{
					if ((i != startId) && (k & parentMask)) 
					{ 
						const Ogre::Plane &sp = getFrustumPlane((unsigned short)i);

						if (sp.getSide(getNearestPoint(i, box)) == Ogre::Plane::NEGATIVE_SIDE) 
						{
							*childrenMask |= k;
							result = PARTIAL;
						}
					} 
				}
				return result;				
			}
			//-----------------------------------------------------------------------

            bool isRegisteredInOcclusionSystem() const;
            void setRegisteredInOcclusionSystem(const bool registered);
            
            void updateRegistrationInOcclusionSystem();

			void setNextOcclusionMode();
            culling_modes getOcclusionMode() const {return mOcclusionMode;};
			bool needRegistrationInOcclusionSystem() const;
			void setOcclusionMode(culling_modes occlusionMode);
			void setOcclusionModeDirect(culling_modes occlusionMode);

			void setOcclusionModeAsString(const Ogre::String &cullingModeAsString);
			Ogre::String getOcclusionModeAsString() const;
			
            inline unsigned int getRenderId() const {return mCurrentRenderUniqueIdentification;}
			void setRenderId(const unsigned int id){mCurrentRenderUniqueIdentification = id;}

            unsigned int getFrameId() const {return mFrameId;}
            bool nextFrame(const unsigned int framesSceneId);

			void _addCHCRenderedFaces(unsigned int numfaces)
			{
				mVisFacesLastCHCRender = numfaces;
			}

			MovableObjectList *getVisiblesList(){return mVisibles;}

			void unregisterInOcclusion();
			void registerInOcclusion();

			void addVisible(MovableObject *m){octree_assert(mVisibles); mVisibles->push_back(m);}

	private:
		void changeOcclusionMode(culling_modes nextOcclusionMode);

		unsigned int mVisFacesLastCHCRender;
        bool isOcclusionSystemRegistered;
        unsigned int mCurrentRenderUniqueIdentification;
        culling_modes mOcclusionMode;
        
        /// last rendered frame
        unsigned int mFrameId;
        unsigned int mFrameSceneId;


		MovableObjectList *mVisibles;

		Ogre::Sphere mFrustumSphere;
		Cone mFrustumCone;



		size_t mPlaneNormalSign[6];

		RadarFrustum mRadar;
	};

}

#endif
