/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/
/***************************************************************************
OctreeCHC.h  -  description
-------------------
begin                : Mon Sep 30 2002
copyright            : (C) 2002 by Jon Anderson
email                : janders@users.sf.net

Enhancements 2003 - 2004 (C) The OGRE Team
***************************************************************************/

#ifndef __OgreSpatialTree_H
#define __OgreSpatialTree_H

#include "OctreeCHCPrerequisites.h"

#include "OctreeCHCOcclusionElement.h"
#include "OgreAxisAlignedBox.h"

#include <list>

namespace OgreOctreeCHC
{
	/** tree data structure for managing scene nodes.
	*/
    class SpatialTree : public OcclusionElement
	{

	    public:
			SpatialTree(SpatialTreeManager *treeManager);

			virtual ~SpatialTree(void);
			virtual void reset();

	public :
			inline void setTreeParent(SpatialTree * parent);
			inline SpatialTree *getTreeParent (); 

			/** Returns the number of scene nodes attached to this tree
		    */
		    inline size_t getNumNodes(void) const; 
			inline bool isLeaf() const;
            inline bool isOccluder () const;
            // check if need to traverse it.
            inline bool hasChildren() const;
			inline size_t getChildrenMaxCount() const;
			
            inline void depthTraversal(DepthTraversal &tr);
            inline void depthTraversal(const DepthTraversalConst &tr);
            inline void depthTraversal(const ConstDepthTraversalConst &tr) const;

			inline void breadthTraversal(BreadthTraversal &tr) ;
			inline void breadthTraversal(const BreadthTraversalConst &tr);
			inline void breadthTraversal(const ConstBreadthTraversalConst &tr) const;



	public :
		    /** Adds an scene node to this tree level.
		    @remarks
			    This is called by the Tree after
			    it has determined the correct tree to insert the node into.
		    */
			virtual void _addNode(SpatialTreeNode* node);
			/** Adds an scene node to this tree level.
			@remarks
				This is called by the SceneManager to
				determine the correct tree to insert the node into.
			*/
			virtual void _insertNode(SpatialTreeNode* node) = 0;			
	        /** Removes an scene node to this tree level.
	        */
	        virtual void _removeNode(SpatialTreeNode* node);		    
		    /** creates the wire frame bounding box for this octant
		    */
			virtual Ogre::WireBoundingBox* getWireBoundingBox(void);		   

		    /** set the Ogre::AxisAlignedBox used for culling this tree.*/
			void setCullBoundingBox(Ogre::AxisAlignedBox *box);
			/** get the Ogre::AxisAlignedBox used for culling this Tree.
		    @remarks
			    if it's a loose spatial indexing, the culling bounds can be different than the actual bounds of the OctreeCHC.
		    */
			inline const Ogre::AxisAlignedBox &_getCullBounds() const;

            virtual void setBoundingBox(const Ogre::Vector3 &min,  const Ogre::Vector3 &max) = 0;

            virtual void setDebugCorners();

			const Ogre::AxisAlignedBox &getBoundingBox() const;

			/** Increments the overall node count of this Tree and all it's parents
			*/
			inline void _ref(void);
			/** Decrements the overall node count of this Tree and all it's parents
			*/
			inline void _unref(void);
			/** Decrements the overall node count of this Tree and all it's parents
			* but may delete some children as this one is empty.
			*/
			void _unref(SpatialTree *oldChild);
	public :
			/** array of children of this Tree.
		    */
			SpatialTree** mChildren;
		    /** Public list of SceneNodes attached to this particular OctreeCHC
		    */
		    SpatialTreeNodeList mNodes;
		    /** Public list of Moving SceneNodes attached to this particular OctreeCHC
		    */
		    SpatialTreeNodeList mMovingNodes;
		    /** Public list of Static SceneNodes attached to this particular OctreeCHC
		    */
		    SpatialTreeNodeList mStaticNodes;

	    protected:

	        /** The bounding box of the tree
		    */
		    Ogre::AxisAlignedBox mBoundingBox;
			/*
			* Bounding box used for culling, might not be the same as mBounding box (Loose spatial indexing)
			*/
		    Ogre::AxisAlignedBox *mCullBoundingBox;
	        /** The Debug visual bounding box of the OctreeCHC
		    */
		    Ogre::WireBoundingBox* mWireBoundingBox;

		public :

		    ///number of SceneNodes in this OctreeCHC and all it's children.
            size_t mNumNodes;
		    ///parent Tree, if any
		    SpatialTree* mTreeParent;
			///The one that handles Trees (new, delete, add, remove, etc...)
            SpatialTreeManager *mTreeManager;
	};

}


#include "OctreeCHCOcclusionTraversal.h"
#include "OgreSpatialTreeManager.h"

namespace OgreOctreeCHC
{
	//-----------------------------------------------------------------------
	void SpatialTree::setTreeParent(SpatialTree * parent) 
	{
		mTreeParent = parent; 
		OcclusionElement::setOcclusionParent(parent);
	};
	//-----------------------------------------------------------------------
	inline SpatialTree* SpatialTree::getTreeParent () {return mTreeParent;};  
	//-----------------------------------------------------------------------
	inline const Ogre::AxisAlignedBox &SpatialTree::getBoundingBox() const {return mBoundingBox;};
	//-----------------------------------------------------------------------
	// check if need to traverse it.
	inline bool SpatialTree::hasChildren() const { return (mNumNodes != 0);};
	//-----------------------------------------------------------------------
	inline size_t SpatialTree::getChildrenMaxCount() const { return mTreeManager->getMaxChildrenCount();};
	//-----------------------------------------------------------------------
	inline bool SpatialTree::isOccluder () const  { return true; }
	//-----------------------------------------------------------------------
	inline bool SpatialTree::isLeaf() const {return false;};
	//-----------------------------------------------------------------------
	inline size_t SpatialTree::getNumNodes(void) const{ return mNumNodes; }; 
	//-----------------------------------------------------------------------
	inline void SpatialTree::depthTraversal(DepthTraversal &tr) {   tr.onDepthTree (*this); };
	//-----------------------------------------------------------------------
	inline void SpatialTree::depthTraversal(const DepthTraversalConst &tr){  tr.onDepthTree(*this); };
	//-----------------------------------------------------------------------
	inline void SpatialTree::depthTraversal(const ConstDepthTraversalConst &tr) const {   tr.onDepthTree (*this); }
	//-----------------------------------------------------------------------
	inline void SpatialTree::breadthTraversal(BreadthTraversal &tr) {   tr.onBreadthTree (*this); };
	//-----------------------------------------------------------------------
	inline void SpatialTree::breadthTraversal(const BreadthTraversalConst &tr){  tr.onBreadthTree(*this); };
	//-----------------------------------------------------------------------
	inline void SpatialTree::breadthTraversal(const ConstBreadthTraversalConst &tr) const {   tr.onBreadthTree (*this); };
	//-----------------------------------------------------------------------
	inline void SpatialTree::_ref(void)
	{
		++mNumNodes;
		if (mTreeParent != 0)
		{
			mTreeParent->_ref();
		}
	};
	//-----------------------------------------------------------------------
	inline void SpatialTree::_unref(void)
	{
		--mNumNodes;
		if (mTreeParent != 0)
			mTreeParent->_unref();
	}
	//-----------------------------------------------------------------------
	inline const Ogre::AxisAlignedBox &SpatialTree::_getCullBounds() const
	{
		return *mCullBoundingBox;
	}
}

#endif //__OgreSpatialTree_H
