/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later__add
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/
/***************************************************************************
OctreeCHC.cpp  -  description
-------------------
begin                : Mon Sep 30 2002
copyright            : (C) 2002 by Jon Anderson
email                : janders@users.sf.net

Enhancements 2003 - 2004 (C) The OGRE Team

***************************************************************************/

#include "OctreeCHCPrecompiledHeaders.h"

#include "OgreWireBoundingBox.h"
#include "OctreeCHCOcclusionBoundingBox.h"

#include "OctreeCHCSceneManager.h"

#include "OgreSpatialTree.h"

#include "OctreeCHC.h"
#include "OctreeCHCNode.h"

#include "OgreStringConverter.h"

using namespace Ogre;

namespace OgreOctreeCHC
{
    //-----------------------------------------------------------------------
    OctreeCHC::OctreeCHC(SpatialTreeManager *treemanager) : 
		   SpatialTree(treemanager),
           mHalfSize(Vector3::ZERO)
    {
		mChildren = reinterpret_cast <SpatialTree**> (new OctreeCHC*[8]);
		//initialize all children to null.
		for (unsigned int i = 0; i < 8; i++)
		{
			mChildren[i] = 0;
        }
    }
    //-----------------------------------------------------------------------
    /** Returns true is the box will fit in a child.
    */
    bool OctreeCHC::_isTwiceSize(const AxisAlignedBox &box) const
	{ 	
		// infinite or null boxes never fit in a child - always root node
		assert (!(box.isInfinite() || box.isNull()));

		const Vector3 &octantHalfBoxSize (mBoundingBox.getHalfSize());
		const Vector3 &boxSize (box.getSize());
		return ((boxSize.x <= octantHalfBoxSize.x) && 
				(boxSize.y <= octantHalfBoxSize.y) && 
				(boxSize.z <= octantHalfBoxSize.z));
    }
    //-----------------------------------------------------------------------
    /** Returns true is the box will fit in a child.
    */
    bool OctreeCHC::_isTwiceCullSize(const AxisAlignedBox &box) const
	{
		// infinite or null boxes never fit in a child - always root node
		assert (!(box.isInfinite() || box.isNull()));

		const Vector3 &boxSize (box.getSize());
		return(boxSize.x  <= mCullHalfSize.x) &&
			(boxSize.y  <= mCullHalfSize.y) &&
			(boxSize.z  <= mCullHalfSize.z) ;
/*
		const Vector3 &octantBoxSize (mBoundingBox.getSize());
		const Vector3 &boxSize (box.getSize());
        return(boxSize.x  <= octantBoxSize.x) &&
              (boxSize.y  <= octantBoxSize.y) &&
              (boxSize.z  <= octantBoxSize.z) ;*/

    }
     //-----------------------------------------------------------------------
    /** Returns true is the box will fit in a child.
    */
    bool OctreeCHC::_isNotCrossingAxes(const AxisAlignedBox &box) const
    { 
        const Vector3 &cullBoxMin (mDoubleBoundingBox.getMinimum());
        const Vector3 &cullBoxMax (mDoubleBoundingBox.getMaximum());

        const Vector3 octCullCenter (cullBoxMin + mCullHalfSize);   

        return !((cullBoxMin.x <= octCullCenter.x && octCullCenter.x <= cullBoxMax.x) ||
                 (cullBoxMin.y <= octCullCenter.y && octCullCenter.y <= cullBoxMax.y) ||
                 (cullBoxMin.z <= octCullCenter.z && octCullCenter.z <= cullBoxMax.z));

    }
    //-----------------------------------------------------------------------
    OctreeCHC *OctreeCHC::_getCullChildWhereBoxFits(const AxisAlignedBox &box)
    { 
        /** It's assumed the the given box has already been proven to fit into
        * a child.  Since it's a loose OctreeCHC, only the centers need to be
        * compared to find the appropriate node.
		*/
		const Vector3 &octantMax (mDoubleBoundingBox.getMaximum());
		const Vector3 &octantMin (mDoubleBoundingBox.getMinimum());

        const Vector3 octCenter (octantMin + mCullHalfSize);   
        const Vector3 ncenter ( box.getMaximum().midPoint(box.getMinimum()));

		octree_assert (octantMax.x >= ncenter.x && octantMax.y >= ncenter.y && octantMax.z >= ncenter.z &&
			octantMin.x <= ncenter.x && octantMin.y <= ncenter.y && octantMin.z <= ncenter.z);

        unsigned int x;
        if (ncenter.x > octCenter.x)
            x = 1;
        else
            x = 0;

        unsigned int y;
        if (ncenter.y > octCenter.y)
            y = 1;
        else
            y = 0;

        unsigned int z;
        if (ncenter.z > octCenter.z)
            z = 1;
        else
            z = 0;

		// index children as 0..7
		const int cid = (x << 2) + (y << 1) + z;
        if (mChildren[ cid ] == 0)
        {
            Vector3 min, max;

            if (x == 0)
            {
                min.x = octantMin.x;
                max.x = octCenter.x;
            }
            else
            {
                min.x = octCenter.x;
                max.x = octantMax.x;
            }

            if (y == 0)
            {
                min.y = octantMin.y;
                max.y = octCenter.y;
            }
            else
            {
                min.y = octCenter.y;
                max.y = octantMax.y;
            }

            if (z == 0)
            {
                min.z = octantMin.z;
                max.z = octCenter.z;
            }
            else
            {
                min.z = octCenter.z;
                max.z = octantMax.z;
            }

#ifdef _DEBUG
			std::cout << "new Static Child\n";
#endif _DEBUG

            SpatialTree *newChild = mTreeManager->getNewTree();
            newChild->setTreeParent (this);
            newChild ->setBoundingBox(min, max);

			octree_assert (newChild -> _getCullBounds().contains(box));

			octree_assert (octantMax.x >= ncenter.x && octantMax.y >= ncenter.y && octantMax.z >= ncenter.z &&
				octantMin.x <= ncenter.x && octantMin.y <= ncenter.y && octantMin.z <= ncenter.z);

            mTreeManager->getSceneManager()->registeredNodeInCamera (newChild);
               
            newChild ->setDebugCorners();
                
            mChildren[ cid ] = newChild;
        }
		return static_cast<OctreeCHC *>(mChildren[ cid ]);
            
    }
    //-----------------------------------------------------------------------
    OctreeCHC *OctreeCHC::_getChildWhereBoxFits(const AxisAlignedBox &box)
    { 
        /** It's assumed the the given box has already been proven to fit into
        * a child.  Since it's a loose OctreeCHC, only the centers need to be
        * compared to find the appropriate node.
        */
        const Vector3 &octantMax (mBoundingBox.getMaximum());
        const Vector3 &octantMin (mBoundingBox.getMinimum());

        const Vector3 octCenter (octantMin + mHalfSize);   
        const Vector3 ncenter (box.getMaximum().midPoint(box.getMinimum()));

        octree_assert (octantMax.x >= ncenter.x && octantMax.y >= ncenter.y && octantMax.z >= ncenter.z &&
                octantMin.x <= ncenter.x && octantMin.y <= ncenter.y && octantMin.z <= ncenter.z);

        unsigned int x;
        if (ncenter.x > octCenter.x)
            x = 1;
        else
            x = 0;

        unsigned int y;
        if (ncenter.y > octCenter.y)
            y = 1;
        else
            y = 0;

        unsigned int z;
        if (ncenter.z > octCenter.z)
            z = 1;
        else
            z = 0;

		const int cid = (x << 2) + (y << 1) + z;
        if (mChildren[ cid ] == 0)
        {
            Vector3 min, max;

            if (x == 0)
            {
                min.x = octantMin.x;
                max.x = octCenter.x;
            }
            else
            {
                min.x = octCenter.x;
                max.x = octantMax.x;
            }

            if (y == 0)
            {
                min.y = octantMin.y;
                max.y = octCenter.y;
            }
            else
            {
                min.y = octCenter.y;
                max.y = octantMax.y;
            }

            if (z == 0)
            {
                min.z = octantMin.z;
                max.z = octCenter.z;
            }
            else
            {
                min.z = octCenter.z;
                max.z = octantMax.z;
            }

#ifdef _DEBUG
            std::cout << "new Child\n";
#endif _DEBUG

            SpatialTree *newChild = mTreeManager->getNewTree();

            newChild->setTreeParent (this);
            newChild->setBoundingBox(min, max);

            octree_assert (max.x >= ncenter.x && max.y >= ncenter.y && max.z >= ncenter.z &&
                    min.x <= ncenter.x && min.y <= ncenter.y && min.z <= ncenter.z);
          

            mTreeManager->getSceneManager()->registeredNodeInCamera (newChild);
               
            newChild ->setDebugCorners();
                
            mChildren[cid] = newChild;
        }        
		return static_cast<OctreeCHC *>(mChildren[ cid ]);
            
    }
    //-----------------------------------------------------------------------
    void OctreeCHC::setBoundingBox(const Vector3 &boxMin, 
                                   const Vector3 &boxMax)
    {
		SpatialTree::setBoundingBox(boxMin, boxMax);
        mCullHalfSize = boxMax - boxMin;
        const Vector3 halfSize (mCullHalfSize * 0.5f);
        mHalfSize = halfSize;
		mDoubleBoundingBox.setExtents(boxMin - halfSize, boxMax + halfSize);

		setCullBoundingBox(&mDoubleBoundingBox);
		setOcclusionBoundingBox(mDoubleBoundingBox);
		//setOcclusionBoundingBox(mBoundingBox);
	}
	//-----------------------------------------------------------------------
	void OctreeCHC::_insertStaticNode(OctreeCHCNode * const n, const int depth)
	{
		const Ogre::AxisAlignedBox &bx (n->_getWorldAABB());

		//if the OctreeCHC is twice as big as the scene node,
		//we will add it to a child.
		if (
			(depth < mTreeManager->getMaxDepth()) 
			&& _isTwiceCullSize (bx)
			)
		{

			if (_isNotCrossingAxes(bx))
			{
				_getCullChildWhereBoxFits(bx)->_insertStaticNode(n, depth + 1);
			}
			else
			{
				// re-insert it as a moving node, therefore avoiding crossing problem.
				n->setStaticCulling (false);
				_insertMovableNode(n, depth);
			}			
		}
		else
		{  

#ifdef _DEBUG
			std::cout << "Depth Static Placement " 
				<< StringConverter::toString (depth) 
				<< " \n";
#endif _DEBUG

			n ->setDebugCorners();
			_addNode(n);
		}
	}
	//-----------------------------------------------------------------------
	void OctreeCHC::_insertMovableNode(OctreeCHCNode * const n, const int depth)
	{
		const Ogre::AxisAlignedBox &bx (n->_getWorldAABB());
		//if the OctreeCHC is twice as big as the scene node,
		//we will add it to a child.
		if (
			(depth < mTreeManager->getMaxDepth()) 
			&& _isTwiceSize (bx)
			)
		{
			_getChildWhereBoxFits(bx)->_insertMovableNode(n, depth + 1);
		}
		else
		{  
			n ->setDebugCorners();


#ifdef _DEBUG
			std::cout << "Depth Placement " 
				<< StringConverter::toString (depth) 
				<< " \n";
#endif _DEBUG

			_addNode(n);
		}
	}
	//-----------------------------------------------------------------------
	void OctreeCHC::_insertNode(SpatialTreeNode* node)
	{
		octree_assert (node);
		if (node->isStaticNode ())
			_insertStaticNode(static_cast<OctreeCHCNode*> (node), 0);
		else
			_insertMovableNode(static_cast<OctreeCHCNode*> (node), 0);
	}
}

