/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/
/***************************************************************************
OctreeCHCcamera.cpp  -  description
-------------------
begin                : Fri Sep 27 2002
copyright            : (C) 2002 by Jon Anderson
email                : janders@users.sf.net

Enhancements 2003 - 2004 (C) The OGRE Team

***************************************************************************/

#include "OctreeCHCPrecompiledHeaders.h"

#include "OgreMath.h"
#include "OgreAxisAlignedBox.h"
#include "OgreRoot.h"
#include "OgreViewport.h"
#include "OgreRenderSystem.h"

#include "OctreeCHCCamera.h"
#include "OctreeCHCSceneManager.h"


#include "OctreeCHCOcclusionCameraTraversal.h"

using namespace Ogre;

namespace OgreOctreeCHC
{
	//-----------------------------------------------------------------------
    OctreeCHCCamera::OctreeCHCCamera(const Ogre::String& name, SceneManager* sm) : 
            Camera(name, sm),
            mVisFacesLastCHCRender(0),
            isOcclusionSystemRegistered (false),
			//mOcclusionMode (VIEW_FRUSTUM_DIRECT),
			mOcclusionMode (CHC),
            mFrameId(0),
			mFrameSceneId(0),
			mCurrentRenderUniqueIdentification (0)   
    {
    }
    //-----------------------------------------------------------------------
    OctreeCHCCamera::~OctreeCHCCamera()
    {
	} 
	//-----------------------------------------------------------------------
	void OctreeCHCCamera::updateCameraPrecalc()
	{
		// update Frustum sphere and Cone to get fast rejection.
		{
			// calculate the radius of the frustum sphere
			const Ogre::Real  fViewLen = mFarDist - mNearDist;
			// use some trig to find the height of the frustum at the far plane
			const Ogre::Real  fHeight = fViewLen * Ogre::Math::Tan(mFOVy * 0.5f);
			// with an aspect ratio of 1, the width will be the same
			const Ogre::Real  fWidth = fHeight*getAspectRatio();
			//set Frustum Sphere
			{

				// halfway point between near/far planes starting at the origin and extending along the z axis
				Ogre::Vector3 P (0.0, 0.0, mNearDist + fViewLen * 0.5f);

				// the calculate far corner of the frustum
				Ogre::Vector3 Q (fWidth, fHeight, fViewLen);

				// the vector between P and Q
				Ogre::Vector3 vDiff (P - Q);

				// the radius becomes the length of this vector
				mFrustumSphere.setRadius(vDiff.length());
				// calculate the center of the sphere
				mFrustumSphere.setCenter(getRealPosition() + (getRealDirection() * (fViewLen * 0.5f) + mNearDist));

			}
			// calculate the length of the fov triangle
			const Ogre::Real fDepth  = fHeight / Ogre::Math::Tan(mFOVy * 0.5f);
			{
				// set the properties of the frustum cone vLookVector is the look vector from the view matrix in the 
				// camera.  Position() returns the position of the camera.  
				// fWidth is half the width of the screen (in pixels).  
				// fHeight is half the height of the screen in pixels. 
				// m_fFovRadians is the FOV of the frustum.



				// calculate the corner of the screen
				const Ogre::Real fCorner = Ogre::Math::Sqrt(fWidth * fWidth + fHeight * fHeight);

				// now calculate the new fov
				const Ogre::Radian fFov = Ogre::Math::ATan(fCorner / fDepth);

				// apply to the cone
				mFrustumCone.setAxis (getRealDirection() );
				mFrustumCone.setVertex (getRealPosition());
				mFrustumCone.setConeAngle (fFov);
			}
		}
		// Radar Approach from Game programming gems 5
		{
			mRadar.SetPerspective(mFOVy, getAspectRatio(), mNearDist, mFarDist);
			mRadar.Build(getRealPosition(), getRealDirection(), getRight(), getUp());
		}
		{
			//plane Masking

			for (unsigned short i = 0; i < 6; i++)
			{ 
				if (i == Ogre::FRUSTUM_PLANE_FAR && getFarClipDistance() == 0)
					continue;

				const Plane &sp = getFrustumPlane(i);

				mPlaneNormalSign[i] = 0;
				if (sp.normal.x > 0) 
					mPlaneNormalSign[i] |= 1;
				if (sp.normal.y > 0) 
					mPlaneNormalSign[i] |= 2;
				if (sp.normal.z > 0) 
					mPlaneNormalSign[i] |= 4;
				
			}
		}
	}
    //-----------------------------------------------------------------------
    bool OctreeCHCCamera::nextFrame(const unsigned int framesSceneId)
    {
        if  (framesSceneId != mFrameSceneId)
        {  
            // change frame Id counter 
            //that identify current frame.
            mFrameSceneId = framesSceneId;
            mFrameId += 1;
            return true;
        }
        return false;
    }
    //----------------------------------------------------------------------- 
    bool OctreeCHCCamera::isRegisteredInOcclusionSystem() const
    {
        return isOcclusionSystemRegistered;
    }
    //-----------------------------------------------------------------------
    void OctreeCHCCamera::setRegisteredInOcclusionSystem(const bool registered)
    {
        isOcclusionSystemRegistered = registered;
	}
	//-----------------------------------------------------------------------
	void OctreeCHCCamera::changeOcclusionMode(culling_modes nextOcclusionMode)
	{
		if (nextOcclusionMode != mOcclusionMode) 
		{
			if (nextOcclusionMode == CHC || 
				nextOcclusionMode == CHC_CONSERVATIVE ||
				nextOcclusionMode == STOP_AND_WAIT
				)
			{  
				if (!getSceneManager ()->getDestinationRenderSystem ()->getCapabilities()->hasCapability(RSC_HWOCCLUSION))
				{
					OGRE_EXCEPT(1, 
						"Your card does not support HWOCCLUSION ", 
						"OctreeCHCCamera::setNextOcclusionMode");
				}
			}

            mOcclusionMode = nextOcclusionMode;
			updateRegistrationInOcclusionSystem();
		}
    }

	//-----------------------------------------------------------------------
	void OctreeCHCCamera::registerInOcclusion()
	{
		static_cast <OctreeCHCSceneManager *> (getSceneManager ())->getTreeManager()->traverse(RegisterCameraDepthTraversal(this));
		setRegisteredInOcclusionSystem (true);		
		mVisibles = new MovableObjectList();
	}
	//-----------------------------------------------------------------------
	void OctreeCHCCamera::unregisterInOcclusion()
	{
		OctreeCHCSceneManager * const scnMgr = static_cast <OctreeCHCSceneManager *> (getSceneManager ());

		if (isRegisteredInOcclusionSystem ())
		{
			scnMgr->getTreeManager()->traverse(UnregisterCameraDepthTraversal(this));
			setRegisteredInOcclusionSystem (false);                   
		} 

		if (scnMgr->getCurrentCurrentOptionCamera() == this)
			scnMgr->setCurrentCurrentOptionCamera(0);

		scnMgr->getTreeManager()->traverse(UnregisterCameraDepthTraversal(this));
		setRegisteredInOcclusionSystem (false);	

		mVisibles->clear();
		delete mVisibles;
		mVisibles = 0;
	}
    //-----------------------------------------------------------------------
    void OctreeCHCCamera::updateRegistrationInOcclusionSystem()
	{
        if (needRegistrationInOcclusionSystem() && !isRegisteredInOcclusionSystem()) 
        {
			registerInOcclusion();
			
        }
        else if (!needRegistrationInOcclusionSystem() && 
				 isRegisteredInOcclusionSystem())
		{	
			unregisterInOcclusion();
        }
		// reset visibility info so that no CHC system use previous frame info
		if (isRegisteredInOcclusionSystem())
			static_cast <OctreeCHCSceneManager *> (getSceneManager ())->getTreeManager()->traverse(ResetCameraVisibilityInfo(this));
    }
    //-----------------------------------------------------------------------
    void OctreeCHCCamera::setNextOcclusionMode()
    {
        culling_modes nextOcclusionMode = static_cast <culling_modes> ((mOcclusionMode + 1) % NUM_CULLING_MODE);
        
        changeOcclusionMode (nextOcclusionMode);
	}
	//-----------------------------------------------------------------------
	bool OctreeCHCCamera::needRegistrationInOcclusionSystem() const
	{		
		// TODO : register only if CHC
		return true;
		
// 		if (mOcclusionMode == CHC || 
// 			mOcclusionMode == CHC_CONSERVATIVE ||
// 			mOcclusionMode == STOP_AND_WAIT)
// 			return true;
// 		else
// 			return false;
	};
	//-----------------------------------------------------------------------
	void OctreeCHCCamera::setOcclusionMode(culling_modes occlusionMode) 
	{
		changeOcclusionMode (occlusionMode);
	};
	//-----------------------------------------------------------------------
	void OctreeCHCCamera::setOcclusionModeDirect(culling_modes occlusionMode) 
	{
		mOcclusionMode  = occlusionMode;
	};
	//-----------------------------------------------------------------------
	void OctreeCHCCamera::setOcclusionModeAsString(const Ogre::String &cullingModeAsString) 
	{
		culling_modes cullingmode (VIEW_FRUSTUM_DIRECT);
		if (cullingModeAsString == "VIEW_FRUSTUM_DIRECT")
			cullingmode = VIEW_FRUSTUM_DIRECT;
		else if (cullingModeAsString == "VIEW_FRUSTUM_MASK")
			cullingmode = VIEW_FRUSTUM_MASK;
		if (cullingModeAsString == "VIEW_FRUSTUM_MASK_TEMPORAL")
			cullingmode = VIEW_FRUSTUM_MASK_TEMPORAL;
		else if (cullingModeAsString == "CHC")
			cullingmode = CHC;
		else if (cullingModeAsString == "CHC_CONSERVATIVE")
			cullingmode = CHC_CONSERVATIVE;
		else if (cullingModeAsString == "VIEW_FRUSTUM")
			cullingmode = VIEW_FRUSTUM;
		else if (cullingModeAsString == "VIEW_FRUSTUM_COMPLEX")
			cullingmode = VIEW_FRUSTUM_COMPLEX;
		else if (cullingModeAsString == "VIEW_FRUSTUM_RADAR")
			cullingmode = VIEW_FRUSTUM_RADAR;
		else if (cullingModeAsString == "STOP_AND_WAIT")
			cullingmode = STOP_AND_WAIT;
		setOcclusionMode(cullingmode);	
	};
	//-----------------------------------------------------------------------
	String OctreeCHCCamera::getOcclusionModeAsString() const
	{
		switch (getOcclusionMode())
		{
		case CHC: return String("CHC");
		case VIEW_FRUSTUM: return String("VIEW_FRUSTUM");
		case VIEW_FRUSTUM_MASK: return String("VIEW_FRUSTUM_MASK");
		case VIEW_FRUSTUM_MASK_TEMPORAL: return String("VIEW_FRUSTUM_MASK_TEMPORAL");
		case VIEW_FRUSTUM_RADAR: return String("VIEW_FRUSTUM_RADAR");
		case VIEW_FRUSTUM_COMPLEX: return String("VIEW_FRUSTUM_COMPLEX");
		case STOP_AND_WAIT: return String("STOP_AND_WAIT");
		case CHC_CONSERVATIVE: return String("CHC_CONSERVATIVE");
		case VIEW_FRUSTUM_DIRECT:return String("VIEW_FRUSTUM_DIRECT");

		default:
			return String("VIEW_FRUSTUM_DIRECT");
		} 
	}
}




