
#include "OctreeCHCPrecompiledHeaders.h"

#include "Ogre.h"

#include "OctreeCHCPrerequisites.h"

#include "OctreeCHCCamera.h"

#include "OctreeCHC.h"
#include "OctreeCHCNode.h"

#include "OctreeCHCOcclusion.h"
#include "OctreeCHCOcclusionElement.h"

#include "OctreeCHCOcclusionSorter.h"
#include "OctreeCHCOcclusionVisibilityData.h"
#include "OctreeCHCOcclusionTraversal.h"
#include "OctreeCHCOcclusionCHCTraversal.h"
#include "OctreeCHCOcclusionSWTraversal.h"
#include "OctreeCHCOcclusionVFTraversal.h"
#include "OctreeCHCOcclusionDebugTraversal.h"

#include "OctreeCHCOcclusionQuerySet.h"

#include "OctreeCHCOcclusionBoundingBox.h"

using namespace Ogre;

namespace OgreOctreeCHC
{
    //-----------------------------------------------------------------------
    OcclusionCulling::OcclusionCulling(unsigned  int visibilityThsd):                
                mVisibilityTreshold(visibilityThsd),
				mIsQueryPoolNotInitiated(true),
#ifdef _DEBUG
				mFrameConservativeVisibility (5)
#else
				mFrameConservativeVisibility (20)
#endif //_DEBUG
    {
    };
    //-----------------------------------------------------------------------
    bool OcclusionCulling::nextFrame(OctreeCHCCamera *cam,  
                                bool onlyshadowcaster,
                                RenderQueue *q)
    {
        bool newframe = true;

        mCurrentRenderQueue = q; 
        mOnlyShadowCaster = onlyshadowcaster;

        if (cam != mCurrentCam)
        {
            mCurrentCam = cam;
        }
        if  (mCurrentCam->nextFrame(Root::getSingleton().getCurrentFrameNumber ()))
        {              
            // change frame Id counter 
            //that identify current frame.
            mFrameId = mCurrentCam->getFrameId();
            newframe = true;     
        }

        if (newframe)
        {
            cam->getVisiblesList()->clear();
            
            // reset counters.
            mTriangleCount = 0;
            mTraversedNodesCount = 0;
            mFrustumCulledNodesCount = 0;
            mQueryCount = 0;
            mObjectCount = 0;
                
            return true;
        }
        return false;
    }
    //-----------------------------------------------------------------------
    void OcclusionCulling::queueDraw(SpatialTreeNode& n) 
    { 
        mObjectCount++; 

        if (!n.notified)
        {   
            // notify objects handled by this node
            n.notifyNodeObjects (mCurrentCam, mOnlyShadowCaster);
        }       
        // as it's visible add its attached objects to render queue.
        // note that we know it already has been notified by camera,
        // when we issued query.        
        n._addAlreadyNotifiedToVisibles ();
        
        // now it had been drawn, notified should be false.
		n.notified = false;

		n.currVis->frameID = mFrameId;         
    }
    //-----------------------------------------------------------------------
    bool OcclusionCulling::issueDrawQuery(SpatialTreeNode& n) 
    {
		mTraversedNodesCount++;
		mObjectCount++;

        octree_assert (n.viewFrustumVisible);
		octree_assert (n.query == 0);
		//octree_assert (n.isOccluder());
		octree_assert (n.currVis);
		octree_assert (n.currVis->frameID != mFrameId) ;


        MovableObjectList *moToRender = n.getVisibleNotifiedNodeObjects(mCurrentCam, mOnlyShadowCaster);
    
        // Mark that Node already notified by this frame/cam combination
        n.currVis->frameID = mFrameId;
        n.notified = true;

        if (!moToRender->empty())
        {
            // create the render Queue.
            MovableObjectList::iterator it = moToRender->begin();
			const MovableObjectList::iterator itend = moToRender->end();
            while (it != itend)
            {
				(*it)->_updateRenderQueue(&*mCurrentRenderQueue);
				if (mIsNotSolidScene)
					mScnMngr->addVisible((*it));
                ++it;
            }

			if (mFirstDrawPending)
			{
				// draw. (no query since nothing was drawn before...)
				mScnMngr->directRenderSingleQueue (&*mCurrentRenderQueue);
				mFirstDrawPending = false;
				moToRender->clear();
				return false;
			}
			else
			{
				mQueryCount++;

				HardwareOcclusionQuery *query =  mQueryPool.getPoolable();

				n.query = query;
				// draw and query.
				query->beginOcclusionQuery ();
				mScnMngr->directRenderSingleQueue (&*mCurrentRenderQueue);
				query->endOcclusionQuery ();
				moToRender->clear();

				return true;
			}
        }
        return false;
    }
    //-----------------------------------------------------------------------
    bool OcclusionCulling::issueBboxQuery(OcclusionElement& n) 
	{
		// query only if something was drawn before...)
		if (mFirstDrawPending)
		{
			return false;
		}

		if (n.getBoundingBox ().intersects(mCurrentCam->getDerivedPosition()))
		{
			// we're inside
			return false;
		}

		mQueryCount++;	       

		octree_assert (n.query == 0);
		HardwareOcclusionQuery *query =  mQueryPool.getPoolable();
		n.query = query;

		query->beginOcclusionQuery();  
			// render bounding box for object 
			mScnMngr->directRenderSingleObject(n.getOcclusionBoundingBox());
		query->endOcclusionQuery();

		//mScnMngr->addVisible (n.getOcclusionBoundingBox());

		return true;
    }
    //-----------------------------------------------------------------------
    bool OcclusionCulling::isQueryResultIsVisible(OcclusionElement& node) 
    {
	     // check visibility result
        unsigned int visiblePixels = 0;
		//this one wait if result not available
        node.query->pullOcclusionQuery (&visiblePixels);
        mQueryPool.removePoolable (node.query);
        node.query = 0;
		if (visiblePixels > mVisibilityTreshold) 
        {
            return true;
        }
        return false;
    }
    //-----------------------------------------------------------------------
    void OcclusionCulling::pullUpVisibility(OcclusionElement& node) 
    {
		OcclusionElement * n = &node;
	    while (n && !n->queriedVisible) 
        {
			n->queriedVisible = true;
			n->currVis->wasQueriedVisible = true;
			n = n->getOcclusionParent();
	    }
    }
    //-----------------------------------------------------------------------
    bool OcclusionCulling::insideViewFrustum(OcclusionElement& n) 
    { 
        const bool result =  mCurrentCam->isVisible (n.getBoundingBox ());
	    	    
        if (result) 
        {
            n.viewFrustumVisible = true;
        }
        else
        {
            n.viewFrustumVisible = false;
	        mFrustumCulledNodesCount++;
        }

	    return result;
    } 
    //-----------------------------------------------------------------------
    //this is the algorithm from the paper
	void OcclusionCulling::CHCdepthTraversal(SpatialTree *tree, 
						Ogre::VisibleObjectsBoundsInfo * const visibleBounds)
    {
        if (mIsQueryPoolNotInitiated)
            initQueryPool ();


		const unsigned int visibilityTreshold = mVisibilityTreshold;

        const Ogre::Vector3 &camPos = mCurrentCam->getDerivedPosition();
	    FrontToBackNodeSorterPriorityQueue depthTraversalStack = FrontToBackNodeSorterPriorityQueue(FrontToBackNodeSorterOperator (camPos));

        //draw leaves, stack otherwise   
        CHCTraversal toStackDepthTraversal(depthTraversalStack, *this, visibleBounds);       
        const unsigned int frameId = mFrameId;
        const unsigned int lastframeId = mFrameId - 1;

        // Octree Root
        depthTraversalStack.push (tree);
		size_t queryWaitCount = 0;
		unsigned int visiblePixels;
	    while (!depthTraversalStack.empty() || !mQueryQueue.empty()) 
		{
			// in case we're waiting a result too long
			if (queryWaitCount > 1000 
				&& !mQueryQueue.empty() 
				&& mQueryQueue.front()->isVisibilityResultReady())
			{
				mQueryQueue.front()->query->pullOcclusionQuery (&visiblePixels);
				queryWaitCount = 0;
			}
			queryWaitCount ++;
			//first part
			// get previous frame occlusion results
 		    while (depthTraversalStack.empty() &&
                !mQueryQueue.empty() && 
                !mQueryQueue.front()->isVisibilityResultReady()) 
            {
			    OcclusionElement& node = *(mQueryQueue.front());
			    mQueryQueue.pop();

				// should be already frustum culled and flagged as not visible
				octree_assert (node.currVis);
                octree_assert (node.currVis->lastQueryFrameId == frameId);
                octree_assert (!node.queriedVisible);
                octree_assert (node.viewFrustumVisible);
                    

                // check visibility result
                visiblePixels = 0;
			    //this one wait if result not available
                node.query->pullOcclusionQuery(&visiblePixels);

				// should store visible pixels per node
				// at last frame for More coherence 
				// at certain threshold 
				// > numVisiblePixelOcluder = occluder only
				// < numVisiblePixelOcludee = test for occlusion only

                mQueryPool.removePoolable(node.query);
                node.query = 0;
			    if (visiblePixels > visibilityTreshold) 
                {
					if (node.isLeaf())
						pullUpVisibility (node);
					else
						pullUpVisibility (node);
                    
				    // Traverse for sure this visible node
                    // And therefore check children
 					node.depthTraversal (toStackDepthTraversal);        
                }
		    }
		    //2nd part
            // query visible objects
            while (!depthTraversalStack.empty()) 
            {
			    OcclusionElement& node = *depthTraversalStack.top ();
			    depthTraversalStack.pop ();

                // make sure there's no loop of some sort.
                // as we must query a node only once 
                if (insideViewFrustum (node))
				{
					octree_assert (node.query == 0);
					node.notified = false;

					VisibilityData * const vis =  node.getNodeData (mCurrentCam);
					octree_assert (frameId != vis->lastQueryFrameId);
					node.currVis = vis;

				    // identify  visible in the last frame nodes
                    const bool wasVisible = vis->wasQueriedVisible && 
                                            lastframeId == vis->lastQueryFrameId;
					if (node.isLeaf())
					{
						if (mFirstDrawPending || (wasVisible && node.isOccluder()))
						{                       
							// Draw & query at the same time

							if (mFirstDrawPending)
								pullUpVisibility (node);

							if (issueDrawQuery (static_cast <SpatialTreeNode &> (node)))
							{
								mQueryQueue.push (&node);
							}
							else
							{
								// skip testing previously visible node
								continue;
							}
						}
						else
						{
							if (issueBboxQuery (node))
								mQueryQueue.push (&node);
						}
					}
					else
					{
						if (mFirstDrawPending || wasVisible)
						{
							// skip testing previously visible node 
							// (interior nodes in hierarchy)
							node.depthTraversal (toStackDepthTraversal);
						}
						else
						{
							// only query node that wasn't visible last frame
							if (issueBboxQuery (node))
								mQueryQueue.push (&node);
							else
								node.depthTraversal (toStackDepthTraversal);    
						}
					}

				    // reset this node's visibility to false
                    // Will be updated by results queries that will pull up results.
                    // by children that will be flagged visible and will transmit info 
					// to their parent nodes (pullUpVisibility())
					node.queriedVisible = false;
				    vis->wasQueriedVisible = false;
				    // update node's visited flag to make sure we won't visit 2 times
                    // and to "tick" the visible info.
				    vis->lastQueryFrameId = frameId;
			    }
		    }
	    }
    }
    //-----------------------------------------------------------------------
    //this is the algorithm from the paper, conservative way.
    void OcclusionCulling::CHCdepthTraversalConservative(SpatialTree *tree, Ogre::VisibleObjectsBoundsInfo * const visibleBounds)
	{
		if (mIsQueryPoolNotInitiated)
			initQueryPool ();

		const unsigned int frameConservativeVisibility = mFrameConservativeVisibility;
		const unsigned int visibilityTreshold = mVisibilityTreshold;

        
        const Ogre::Vector3 &camPos = mCurrentCam->getDerivedPosition();
	    FrontToBackNodeSorterPriorityQueue depthTraversalStack = FrontToBackNodeSorterPriorityQueue(FrontToBackNodeSorterOperator (camPos));

        //draw leaves, stack otherwise   
        CHCTraversal toStackDepthTraversal(depthTraversalStack, *this, visibleBounds);       
        const unsigned int frameId = mFrameId;
        const unsigned int lastframeId = mFrameId - 1;

        // Octree Root
		depthTraversalStack.push (tree);
		size_t queryWaitCount = 0;
		unsigned int visiblePixels;
		while (!depthTraversalStack.empty() || !mQueryQueue.empty()) 
		{
			// in case we're waiting a result too long
			if (queryWaitCount > 1000 
				&& !mQueryQueue.empty() 
				&& mQueryQueue.front()->isVisibilityResultReady())
			{
				mQueryQueue.front()->query->pullOcclusionQuery (&visiblePixels);
				queryWaitCount = 0;
			}
			queryWaitCount ++;
		    //first part
            // get previous frame occlusion results
 		    while (depthTraversalStack.empty() &&
                !mQueryQueue.empty() && 
                !mQueryQueue.front()->isVisibilityResultReady()) 
            {
			    OcclusionElement& node = *(mQueryQueue.front());
			    mQueryQueue.pop();

				// should be already frustum culled and flagged as not visible
				octree_assert (node.currVis);
                octree_assert (node.currVis->lastQueryFrameId == frameId);
                octree_assert (!node.queriedVisible);
                octree_assert (node.viewFrustumVisible);                    

                // check visibility result
                visiblePixels = 0;
			    //this one wait if result not available
				node.query->pullOcclusionQuery(&visiblePixels);

				// should store visible pixels per node
				// at last frame for More coherency 
				// at certain threshold 
				// > numVisiblePixelOcluder = occluder only
				// < numVisiblePixelOcludee = test for occlusion only

                mQueryPool.removePoolable(node.query);
                node.query = 0;
			    if (visiblePixels > visibilityTreshold) 
                {                    
				    // Traverse for sure this visible node
                    // And therefore check children
 					node.depthTraversal (toStackDepthTraversal);    
					pullUpVisibility (node);
                }
		    }
		    //2nd part
            // query visible objects
            while (!depthTraversalStack.empty()) 
            {
			    OcclusionElement& node = *depthTraversalStack.top ();
			    depthTraversalStack.pop ();

                // make sure there's no loop of some sort.
                // as we must query a node only once 
                if (insideViewFrustum (node)) 
				{
					octree_assert (node.query == 0);
					node.notified = false;

					VisibilityData * const vis =  node.getNodeData (mCurrentCam);
					octree_assert (frameId != vis->lastQueryFrameId);
					node.currVis = vis;

				    // identify  visible in the last frame nodes
                    const bool wasVisible = vis->wasQueriedVisible;
                    const bool wasVisibleandQueriedBeforeLastNFrame = wasVisible 
                                                                    && (lastframeId - frameConservativeVisibility < vis->lastQueryFrameId);
                    if (mFirstDrawPending || wasVisibleandQueriedBeforeLastNFrame)
                    {
						// Conservative Case => as was visible many frame... it still is.

						// was visible last frame and still didn't draw anything
                        if (node.isLeaf() )
						{ 
							if (node.isOccluder())
							{
								// drew for sure as it's leaf and an occluder
								// queried only if not first draw pending.

								// dont query if visible due to conservative reasons.
								mObjectCount++;
								MovableObjectList *moToRender = static_cast <SpatialTreeNode *> 
									(&node)->getVisibleNotifiedNodeObjects(mCurrentCam, mOnlyShadowCaster);
								// create the render Queue.
								if (!moToRender->empty())
								{
									MovableObjectList::iterator it = moToRender->begin(), 
																itend = moToRender->end();
									while (it != itend)
									{
										(*it)->_updateRenderQueue (&*mCurrentRenderQueue);
										++it;
									}
									mScnMngr->directRenderSingleQueue (&*mCurrentRenderQueue);
									mFirstDrawPending = false;
								}
								node.notified = true;
							}
							// as it "wasVisibleandQueriedBeforeLastNFrame" state it as visible.
							pullUpVisibility (node);
						}
						else
						{
							// as it "wasVisibleandQueriedBeforeLastNFrame" consider it as visible.
							// and test children
							node.depthTraversal (toStackDepthTraversal);
						}
                    }
                    else // Exactly like CHC standard Case 
                    {
                        const bool wasVisibleandQueriedExacltyLastNFrame = wasVisible && 
                                                                           lastframeId - frameConservativeVisibility == vis->lastQueryFrameId;
                        if (node.isLeaf())
                        {
                            if (mFirstDrawPending || (wasVisibleandQueriedExacltyLastNFrame && node.isOccluder()))
                            {              
								if (mFirstDrawPending)
									pullUpVisibility (node); 

				                // Draw & query at the same time previously visible nodes
                                if (issueDrawQuery (static_cast <SpatialTreeNode &> (node)))
                                {
                                    mQueryQueue.push (&node);
                                }
                                else
                                {
									// skip testing previously visible node
									continue;
                                }
                            }
                            else
							{
								if (issueBboxQuery (node))
									mQueryQueue.push (&node);
                            }
                        }
                        else
                        {
                            if (mFirstDrawPending || wasVisibleandQueriedExacltyLastNFrame)
                            {
                                // skip testing previously visible node 
                                // (interior nodes in hierarchy)
								node.depthTraversal (toStackDepthTraversal);
				            }
                            else
                            {
								// only query node that wasn't visible last frame
								if (issueBboxQuery (node))
									mQueryQueue.push (&node);
								else
									node.depthTraversal (toStackDepthTraversal);     
				            }
                        }
				        // reset this node's visibility to false
                        // Will be updated by results queries that will pull up results.
                        // by children that will be flagged visible and will transmit info 
						// to their parent nodes (pullUpVisibility())
						node.queriedVisible = false;
						vis->wasQueriedVisible = false;
				        // update node's visited flag to make sure we won't visit 2 times
                        // and to "tick" the visible info.
				        vis->lastQueryFrameId = frameId;
                    }
			    }
		    }
	    }
    }
}