/*
-----------------------------------------------------------------------------
This source file is part of OGRE
    (Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#include "OctreeCHCPrecompiledHeaders.h"

#include "Ogre.h"
#include "OctreeCHCOcclusionBoundingBox.h"

#include "OgreSimpleRenderable.h"
#include "OgreHardwareBufferManager.h"
#include "OgreCamera.h"


using namespace Ogre;

namespace OgreOctreeCHC {
    #define POSITION_BINDING 0

    //-----------------------------------------------------------------------
	OcclusionBoundingBox::OcclusionBoundingBox() : 
		SimpleRenderable()
    {
        mRenderOp.indexData = new IndexData();
        mRenderOp.vertexData = new VertexData();    
        mRenderOp.operationType = RenderOperation::OT_TRIANGLE_LIST;
        mRenderOp.indexData->indexCount = 36;
		mRenderOp.vertexData->vertexCount = 8; 
		mRenderOp.vertexData->vertexStart = 0; 
		mRenderOp.useIndexes = true; 

        VertexDeclaration* decl = mRenderOp.vertexData->vertexDeclaration;
        VertexBufferBinding* bind = mRenderOp.vertexData->vertexBufferBinding;

        decl->addElement(POSITION_BINDING, 0, VET_FLOAT3, VES_POSITION);
        

        mRenderOp.indexData->indexBuffer = HardwareBufferManager::getSingleton().createIndexBuffer(
                            HardwareIndexBuffer::IT_16BIT,
                            mRenderOp.indexData->indexCount, 
                            HardwareBuffer::HBU_STATIC_WRITE_ONLY);

        HardwareVertexBufferSharedPtr vbuf = 
            HardwareBufferManager::getSingleton().createVertexBuffer(
                decl->getVertexSize(POSITION_BINDING),
                mRenderOp.vertexData->vertexCount,
                HardwareBuffer::HBU_STATIC_WRITE_ONLY);

        // Bind buffer
        bind->setBinding(POSITION_BINDING, vbuf);

        this->setPolygonModeOverrideable (false);
        //Renderable::getPolygonMode ();

        // set basic white material
		this->setMaterial("CHCSolidOcclusion");
		//this->setMaterial("WhiteOverlayInfo");
	}
    //-----------------------------------------------------------------------
    PolygonMode OcclusionBoundingBox:: getRenderDetail() const 
    {
        return PM_SOLID;
    } 

    //-----------------------------------------------------------------------
    OcclusionBoundingBox::~OcclusionBoundingBox() 
    {
        delete mRenderOp.indexData;
        delete mRenderOp.vertexData;
	}
    //-----------------------------------------------------------------------
    void OcclusionBoundingBox::setupBoundingBox(const Ogre::AxisAlignedBox& aabb) 
    {
		if (aabb.isFinite())
		{
			// init the vertices to the aabb
			setupBoundingBoxVertices(aabb);

    		// setup the bounding box of this SimpleRenderable
			SimpleRenderable::setBoundingBox(aabb);
			
			mBoundingSphere.setCenter(aabb.getCenter());
			mBoundingSphere.setRadius(mRadius);
		}
	}
    //-----------------------------------------------------------------------
    // Override this method to prevent parent transforms (rotation,translation,scale)
    void OcclusionBoundingBox::getWorldTransforms(Matrix4* xform) const
    {
		// return identity matrix to prevent parent transforms
        *xform = Matrix4::IDENTITY;
    }
    //-----------------------------------------------------------------------
    const Quaternion& OcclusionBoundingBox::getWorldOrientation(void) const
    {
        return Quaternion::IDENTITY;
    }
    //-----------------------------------------------------------------------
    const Ogre::Vector3& OcclusionBoundingBox::getWorldPosition(void) const
    {
        return Ogre::Vector3::ZERO;
    }
    //-----------------------------------------------------------------------
    void OcclusionBoundingBox::setupBoundingBoxVertices(const Ogre::AxisAlignedBox& aab) 
    {
		const Ogre::Vector3 &max = aab.getMaximum();
		const Ogre::Vector3 &min = aab.getMinimum();

        const Real sqLen = std::max(max.squaredLength(), 
                                    min.squaredLength());
        mRadius = Math::Sqrt(sqLen);
		
		
        VertexDeclaration * const decl = mRenderOp.vertexData->vertexDeclaration;
        const VertexElement* poselem = decl->findElementBySemantic(VES_POSITION);

        
        HardwareVertexBufferSharedPtr vbuf =
            mRenderOp.vertexData->vertexBufferBinding->getBuffer(POSITION_BINDING);     
           
        const size_t vertexSize = vbuf->getVertexSize ();
        float *pPos;

        uchar* pMain = static_cast<uchar *>(
            vbuf->lock(HardwareBuffer::HBL_DISCARD));

     
		    //generate 8 corners of the bbox
            // RightHanded
		
            #define V3(A_X, A_Y, A_Z, ACOLOR) poselem->baseVertexPointerToElement(pMain, &pPos); \
                    *pPos++ = static_cast <float> (A_X); \
					*pPos++ = static_cast <float> (A_Y); \
					*pPos++ = static_cast <float> (A_Z); \
                    pMain += vertexSize;
			/*
						V3(min.x, max.y, max.z, ColourValue::White) // 1 
						V3(min.x, min.y, max.z, ColourValue::White) // 2 
						V3(max.x, max.y, max.z, ColourValue::White) // 3 
						V3(max.x, min.y, max.z, ColourValue::White) // 4 
						V3(max.x, max.y, min.z, ColourValue::White) // 5 
						V3(max.x, min.y, min.z, ColourValue::White) // 6 
						V3(min.x, max.y, min.z, ColourValue::White) // 7 
						V3(min.x, min.y, min.z, ColourValue::White) // 8
			*/

			V3(min.x,max.y,min.z, ColourValue::White)      //0 position
			V3(max.x,max.y,min.z, ColourValue::White)      //1 position
			V3(max.x,min.y,min.z, ColourValue::White)      //2 position
			V3(min.x,min.y,min.z, ColourValue::White)      //3 position
			V3(min.x,max.y,max.z, ColourValue::White)      //4 position
			V3(max.x,max.y,max.z, ColourValue::White)      //5 position
			V3(max.x,min.y,max.z, ColourValue::White)      //6 position
			V3(min.x,min.y,max.z, ColourValue::White)      //7 position

        vbuf->unlock();

		
        HardwareIndexBufferSharedPtr iBuf = mRenderOp.indexData->indexBuffer;     
        ushort* pIdx = static_cast<ushort*>(
			iBuf->lock(0, iBuf->getSizeInBytes(),HardwareBuffer::HBL_DISCARD));

#define I3(I_X, I_Y, I_Z) {*pIdx++ = static_cast<ushort> (I_X); *pIdx++ = static_cast<ushort> (I_Y); *pIdx++ = static_cast<ushort> (I_Z);}

			/*
					I3(0,1,2);
					I3(1,3,2);
					I3(2,3,4);		
					I3(3,5,4);		
					I3(6,5,7);		
					I3(5,7,6);		
					I3(6,7,0);		
					I3(7,1,0);		
					I3(4,6,0);		
					I3(4,0,2);		
					I3(1,7,3);
					I3(7,5,3);
			*/ 	
			I3(0,2,3);
			I3(1,6,2);
			I3(1,5,6);
			I3(4,6,5);
			I3(4,7,6);
			I3(0,7,4);
			I3(0,3,7);
			I3(0,5,1);
			I3(0,4,5);
			I3(2,7,3);
			I3(2,6,7);
        iBuf->unlock();       
	}
}

