/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/
/***************************************************************************
OctreeCHCscenemanager.cpp  -  description
-------------------
begin                : Fri Sep 27 2002
copyright            : (C) 2002 by Jon Anderson
email                : janders@users.sf.net
 
Enhancements 2003 - 2004 (C) The OGRE Team
 
***************************************************************************/

#include "OctreeCHCPrecompiledHeaders.h"

#include "OctreeCHCSceneManager.h"

#include "OctreeCHCAxisAlignedBoxSceneQuery.h"
#include "OctreeCHCIntersectionSceneQuery.h"
#include "OctreeCHCPlaneBoundedVolumeListSceneQuery.h"
#include "OctreeCHCRaySceneQuery.h"
#include "OctreeCHCSphereSceneQuery.h"
#include "OgreRenderSystem.h"
#include "OgreWireBoundingBox.h"
#include "OgreRenderOperation.h"
#include "OgreStringConverter.h"
#include "OgreLogManager.h"

#include "OgreOverlayManager.h"
#include "OgreOverlayElement.h"

#include "OctreeCHCCamera.h"
#include "OctreeCHC.h"
#include "OctreeCHCNode.h"

#include "OctreeCHCOcclusionBoundingBox.h"
#include "OctreeCHCOcclusionSorter.h"
#include "OctreeCHCOcclusionVisibilityData.h"
#include "OctreeCHCOcclusionTraversal.h"
#include "OctreeCHCOcclusionCameraTraversal.h"
#include "OctreeCHCOcclusionVFTraversal.h"
#include "OctreeCHCOcclusionSWTraversal.h"
#include "OctreeCHCOcclusionDebugTraversal.h"

#include "OctreeCHCManager.h"

using namespace Ogre;

namespace OgreOctreeCHC
{
	int OctreeCHCSceneManager::intersect_call = 0;

    //-----------------------------------------------------------------------
    //-----------------------------------------------------------------------
    //-----------------------------------------------------------------------
    const Ogre::String OctreeCHCSceneManagerFactory::FACTORY_TYPE_NAME = "OctreeCHCSceneManager";
    //-----------------------------------------------------------------------
    void OctreeCHCSceneManagerFactory::initMetaData(void) const
    {
       mMetaData.typeName = FACTORY_TYPE_NAME;

       mMetaData.description = "Scene manager organising the scene on the basis of an octree, possibly using occlusion culling.";
	   mMetaData.sceneTypeMask = 0xFFFF; // support all types
	   mMetaData.worldGeometrySupported = false;
	
	}
	//-----------------------------------------------------------------------
	static void debugMaterialCreate(const String &name, const ColourValue &color, Real alphaVal)
	{
		MaterialManager * const mMgr = MaterialManager::getSingletonPtr();
		MaterialPtr debug = mMgr->create(name, "TreeDebug");
		Pass * const p = debug->createTechnique()->createPass();

		p->setSceneBlending(SBT_TRANSPARENT_ALPHA);
		p->setDepthWriteEnabled(false);
		p->setLightingEnabled(false);
		p->setPolygonMode(PM_WIREFRAME);

		TextureUnitState * const tu = p->createTextureUnitState();
		tu->setColourOperationEx(LBX_SOURCE1, 
			LBS_MANUAL, LBS_CURRENT,
			color);
		tu->setAlphaOperation(LBX_SOURCE1, 
			LBS_MANUAL, LBS_CURRENT,
			alphaVal);

	}
    //-----------------------------------------------------------------------
    SceneManager* OctreeCHCSceneManagerFactory::createInstance(
        const Ogre::String& instanceName)
    {
		Ogre::LogManager::getSingleton().getDefaultLog()->logMessage("OctreeCHCSceneManager Initialisation: " + instanceName);

		{
			MaterialManager *mMgr = MaterialManager::getSingletonPtr();
			if (!mMgr->resourceExists("CHCSolidOcclusion"))
			{
				/*	CHCSolidOcclusion : needed to check if Bounding boxes is visible
				using Hardware Queries.
				*/
				MaterialPtr occlusionMat = mMgr->create("CHCSolidOcclusion", "CHC");
				Pass * const p = occlusionMat->createTechnique()->createPass();

				p->setSceneBlending (SBT_TRANSPARENT_ALPHA);
				p->setDepthWriteEnabled (false);
				p->setLightingEnabled (false);
				p->setColourWriteEnabled (false);
				p->setPolygonMode (PM_SOLID);
				p->setCullingMode (CULL_NONE);
				//p->setCullingMode(CullingMode::CULL_CLOCKWISE);
			}

			/*
			used for debugging purpose...
			*/
			if (!mMgr->resourceExists("WhiteOverlayInfo"))
				debugMaterialCreate("WhiteOverlayInfo", ColourValue::White,  0.5);
			if (!mMgr->resourceExists("BlackOverlayInfo"))
				debugMaterialCreate("BlackOverlayInfo", ColourValue::Black,  0.5);
			if (!mMgr->resourceExists("RedOverlayInfo"))
				debugMaterialCreate("RedOverlayInfo"  , ColourValue::Red,    0.5);
			if (!mMgr->resourceExists("GreenOverlayInfo"))
				debugMaterialCreate("GreenOverlayInfo", ColourValue::Green,  0.5);
			if (!mMgr->resourceExists("BlueOverlayInfo"))
				debugMaterialCreate("BlueOverlayInfo" , ColourValue::Blue,   0.5);

		}

		SceneManager* scnMgr = new OctreeCHCSceneManager(instanceName);

		Ogre::LogManager::getSingleton().getDefaultLog()->logMessage("OctreeCHCSceneManager Created: " + instanceName);
     

		return scnMgr; 
    }
    //-----------------------------------------------------------------------
    void OctreeCHCSceneManagerFactory::destroyInstance(Ogre::SceneManager* instance)
    {
        delete instance;
    }
    //---------------------------------------------------------------------
    OctreeCHCSceneManager::OctreeCHCSceneManager(const Ogre::String& name): 
        SceneManager(name),
		mTreeManager(0),
		mCurrentOptionCamera(0),
		mCurrentSpatialManagerType(LOOSE_OCTREE),
		mRenderTemporalIdentificationIdCount(0)
    {
        init(
			Ogre::AxisAlignedBox(-500, -500, -500, 
								  500,  500,  500), 
			64
			);
    }
    //---------------------------------------------------------------------
    OctreeCHCSceneManager::OctreeCHCSceneManager(const Ogre::String& name, Ogre::AxisAlignedBox &box, int max_depth) : 
        SceneManager(name),
        mTreeManager(0),
		mCurrentOptionCamera(0),
		mCurrentSpatialManagerType(LOOSE_OCTREE),
		mRenderTemporalIdentificationIdCount(0)
    {
        init(box, max_depth);
    }
    //---------------------------------------------------------------------
    const Ogre::String& OctreeCHCSceneManager::getTypeName(void) const
    {
     	return OctreeCHCSceneManagerFactory::FACTORY_TYPE_NAME;
    }
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::init(const Ogre::AxisAlignedBox &box, int depth)
    {
        delete mSceneRoot; //get rid of old root.

        // -- Changes by Steve
        // Don't do it this way, it will add it to the mSceneNodes which we don't want
        //mSceneRoot = createSceneNode("SceneRoot");
        mSceneRoot = new OctreeCHCNode(this, "SceneRoot");
	    mSceneRoot->_notifyRootNode();
        // -- End changes by Steve


        if (mTreeManager)
            delete mTreeManager;

		switch(mCurrentSpatialManagerType)
		{
			case LOOSE_OCTREE:
				mTreeManager = new OctreeCHCManager(this, box, depth);
				break;

			case ABT_OCTREE:
			case SPHERE_OCTREE:
			default:
				mTreeManager = new OctreeCHCManager(this, box, depth);
				break;
		}
		mTreeManager->resize(box);
		//mTreeManager->_insertNode(static_cast<SpatialTreeNode *>(mSceneRoot));

            mShowBoxes = false;
            mCullCamera = false;
			mCullDebug = false;
			mCullDebugText = false;
           //     setDisplaySceneNodes(true);
        
        mCulling.init (this);
        mOcclusionDepth = 0;

    }
    //---------------------------------------------------------------------
    OctreeCHCSceneManager::~OctreeCHCSceneManager()
    {

        // -- Changed by Steve
        // Don't do this here, SceneManager will do it
        /*
        if(mSceneRoot)
        delete mSceneRoot;
        */ 
        // --End Changes by Steve


		CameraList::iterator i = mCameras.begin();
		for (; i != mCameras.end(); ++i)
		{
			static_cast <OctreeCHCCamera *> (i->second)->unregisterInOcclusion();		
		}
		mCamInProgressVisibles = 0;


		std::vector <RenderTemporalIdentification *>::const_iterator it = mRenderTemporalIdentifications.begin();
		for (;it != mRenderTemporalIdentifications.end(); ++it)
		{
			delete *it;
		}
		mRenderTemporalIdentifications.clear();


        if (mTreeManager)
        {
            delete mTreeManager;
            mTreeManager = 0;
        }
    }
    //---------------------------------------------------------------------
    Camera * OctreeCHCSceneManager::createCamera(const Ogre::String &name)
    {
        if (mCameras.find(name) != mCameras.end())
        {
            OGRE_EXCEPT(
                Exception::ERR_DUPLICATE_ITEM,
                "A camera with the name " + name + " already exists",
                "OctreeCHCSceneManager::createCamera");
        }
        Camera * c = new OctreeCHCCamera(name, this);
        mCameras.insert(CameraList::value_type(name, c));
		addCamera (c);

		// create visible bounds aab map entry
		mCamVisibleObjectsMap[c] = VisibleObjectsBoundsInfo();
        return c;
    }
    //-----------------------------------------------------------------------
    void OctreeCHCSceneManager::addCamera(Ogre::Camera *cam)
	{
		OctreeCHCCamera *c = static_cast <OctreeCHCCamera *> (cam);
        if (!c->isRegisteredInOcclusionSystem()
			&& c->needRegistrationInOcclusionSystem())
        {
            c->registerInOcclusion();
        }
		if (!mCurrentOptionCamera)
			mCurrentOptionCamera = c;
	}
    //-----------------------------------------------------------------------
    void OctreeCHCSceneManager::destroyCamera(Ogre::Camera *cam)
    {
        static_cast <OctreeCHCCamera *> (cam)->unregisterInOcclusion();
        SceneManager::destroyCamera(cam);
    }
    //-----------------------------------------------------------------------
    void OctreeCHCSceneManager::destroyCamera(const Ogre::String& name)
    {
		// Find in list
       CameraList::iterator i = mCameras.find(name);
       if (i != mCameras.end())
	   {
          destroyCamera (i->second);
       }                  
    }
    //-----------------------------------------------------------------------
    void OctreeCHCSceneManager::destroyAllCameras(void)
    {
		CameraList::iterator i = mCameras.begin();
		for (; i != mCameras.end(); ++i)
		{
			static_cast <OctreeCHCCamera *> (i->second)->unregisterInOcclusion();
			// Notify render system
			mDestRenderSystem->_notifyCameraRemoved(i->second);
			delete i->second;
		}
        mCameras.clear ();
    } 
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::destroySceneNode(const Ogre::String &name)
    {
        SceneNode * on = static_cast < OctreeCHCNode* > (Ogre::SceneManager::getSceneNode(name));

        if (on != 0)
			mTreeManager->_removeNode(static_cast <SpatialTreeNode *>(on));
        SceneManager::destroySceneNode(name);
    }
    //---------------------------------------------------------------------
    bool OctreeCHCSceneManager::getOptionValues(const Ogre::String & key, StringVector  &refValueList)
    {
        return SceneManager::getOptionValues(key, refValueList);
    }
    //---------------------------------------------------------------------
    bool OctreeCHCSceneManager::getOptionKeys(StringVector & refKeys)
    {
        SceneManager::getOptionKeys(refKeys);

        refKeys.push_back("CullCamera");
        refKeys.push_back("Size");
        refKeys.push_back("ShowOctree");
        refKeys.push_back("Depth");

		refKeys.push_back("setCullingMode");
		refKeys.push_back("CullDebug");
		refKeys.push_back("DebugText");

        return true;
    }
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::_updateSceneNode(SceneNode * node)
    {
		mTreeManager->_updateNode(static_cast <SpatialTreeNode *>(node));
    }
    //---------------------------------------------------------------------
    /** Only removes the node from the OctreeCHC.  It leaves the OctreeCHC, even if it's empty.
    */
    void OctreeCHCSceneManager::_removeSceneNode(SceneNode * node)
    {
        octree_assert (node);
        // Skip if TreeManager has been destroyed (shutdown conditions)
        if (!mTreeManager)
            return;
		mTreeManager->_removeNode(static_cast <SpatialTreeNode *>(node));
		unregisteredNodeInCamera (dynamic_cast <OcclusionElement *>(node));
    }
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::_addSceneNode(SceneNode * n, SpatialTree *tree, const int depth)
    {
        octree_assert (n && tree);
		mTreeManager->_insertNode(static_cast <SpatialTreeNode *>(n));
		registeredNodeInCamera (dynamic_cast <OcclusionElement *>(n));
    }
    //---------------------------------------------------------------------
    SceneNode * OctreeCHCSceneManager::createSceneNode(void)
	{
        SceneNode * on = mTreeManager->getNewTreeNode();
		mSceneNodes[ on->getName() ] = on; 
        return on;
    }
    //---------------------------------------------------------------------
    SceneNode * OctreeCHCSceneManager::createSceneNode(const Ogre::String &name)
    {
		// Check name not used
	    if (mSceneNodes.find(name) != mSceneNodes.end())
	    {
		    OGRE_EXCEPT(
			        Exception::ERR_DUPLICATE_ITEM,
			        "A scene node with the name " + name + " already exists",
			        "OctreeCHCSceneManager::createSceneNode" );
		}
        SceneNode * on = mTreeManager->getNewTreeNode(name);
		mSceneNodes[ on->getName() ] = on; 
        return on;
    }
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::registeredNodeInCamera(OcclusionElement *on)
    {
         // register existing camera in nodes
        bool isRegistered = false;
        for (CameraList::iterator it = mCameras.begin(); it != mCameras.end(); ++it)
        {
            // Notify occlusion system
            OctreeCHCCamera *c = static_cast <OctreeCHCCamera *> (it->second);
            octree_assert (c);
            if (c->isRegisteredInOcclusionSystem ())
            {
                on->depthTraversal (RegisterCameraDepthTraversal(c));
                isRegistered = true;
            }
        }
        on->setRegisteredtoCam (isRegistered);      
	}
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::unregisteredNodeInCamera(OcclusionElement *on)
    {
         // register existing camera in nodes
        for (CameraList::iterator it = mCameras.begin();it != mCameras.end(); ++it)
        {
            // Notify occlusion system
            OctreeCHCCamera *c = static_cast <OctreeCHCCamera *> (it->second);
            octree_assert (c);
            if (c->isRegisteredInOcclusionSystem ())
                on->depthTraversal (UnregisterCameraDepthTraversal(c));
        }   
        on->setRegisteredtoCam (false);               
	}
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::enableHardwareOcclusionTests()
    { 
        Camera * const c_cam = mCameraInProgress;

		c_cam->getViewport()->setClearEveryFrame (false);
        // make sure we always occlude in SOLID MODE
		mBeforeHWOCCamDetail =  c_cam->getPolygonMode ();
		if (mBeforeHWOCCamDetail != PM_SOLID)
		{			
			mCulling.setIfNotSolidScene	(true);
			c_cam->setPolygonMode (PM_SOLID);
		}
		else 
		{
			mCulling.setIfNotSolidScene	(false);
		}

//#define HWOCCLUSIONRTT
#ifdef HWOCCLUSIONRTT

			const Viewport * const c_camViewPort = c_cam->getViewport();
			octree_assert (c_camViewPort);
            if (mOcclusionDepth == 0)
            {
                mOcclusionDepth = mDestRenderSystem->
                    createRenderTexture(
                    "OcclusionDepthTexture", 
                    c_camViewPort->getActualWidth() / 8, 
                    c_camViewPort->getActualHeight() / 8, 
                    TEX_TYPE_2D, 
                    PF_DEPTH);


                mOcclusionDepth->setAutoUpdated (false);
                mOcclusionDepth->setActive (false);

                mOcclusionCamera = createCamera ("OcclusionDepthTextureCamera");
                Viewport * const v = mOcclusionDepth->addViewport(mOcclusionCamera);
				octree_assert (v);
                v->setOverlaysEnabled (false);
                v->setClearEveryFrame(false);
                v->setBackgroundColour(ColourValue::Black);

#define RENDERCOLOR
#ifdef RENDERCOLOR
                    MaterialPtr mat = MaterialManager::getSingleton().create (
                                        "OcclusionDepthTextureMat",
                                        ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);
                    mat->getTechnique(0)->getPass(0)->createTextureUnitState("OcclusionDepthTexture");

                    Overlay *OcclusionDepthOverlay = OverlayManager::getSingleton ().create("OcclusionDepthOverlay"); 
                    OverlayElement *OcclusionDepthPanel = 
                    OverlayManager::getSingleton ().createOverlayElement ("Panel", 
                                                                            "OcclusionDepthPanel", 
                                                                            false); 
                    OcclusionDepthPanel->setMaterialName ("OcclusionDepthTextureMat"); 
                    OcclusionDepthOverlay->add2D ((OverlayContainer *)OcclusionDepthPanel); 
                    OcclusionDepthOverlay->show ();

                    //OcclusionDepthOverlay->setScale (0.5f, 0.5f);
                    OcclusionDepthPanel->setPosition (0.8, 0.8);
#endif //RENDERCOLOR

                mOcclusionCamera->setNearClipDistance(c_cam->getNearClipDistance());
                mOcclusionCamera->setFarClipDistance(c_cam->getFarClipDistance());
                mOcclusionCamera->setAspectRatio(
                    (Real)c_camViewPort->getActualWidth() /  c_camViewPort->getActualHeight());
            }

            // copy camera characteristics to occlusion camera.
            mOcclusionCamera->setPosition (c_cam->getWorldPosition ());
            mOcclusionCamera->setOrientation (c_cam->getWorldOrientation ());
            mOcclusionCamera->setNearClipDistance(c_cam->getNearClipDistance());
            mOcclusionCamera->setFarClipDistance(c_cam->getFarClipDistance());  


            mDestRenderSystem->_setViewport(mOcclusionCamera->getViewport ());
            mDestRenderSystem->_setWorldMatrix(Matrix4::IDENTITY);
            mDestRenderSystem->_setViewMatrix(c_cam->getViewMatrix());
            mDestRenderSystem->_setProjectionMatrix(c_cam->getProjectionMatrix());

            mDestRenderSystem->_beginFrame();

            mDestRenderSystem->clearFrameBuffer (FBT_DEPTH | FBT_COLOUR, 
            mOcclusionCamera->getViewport ()->getBackgroundColour ());

        #else //HWOCCLUSIONRTT

			octree_assert (c_cam->getViewport());
            mDestRenderSystem->_setViewport(c_cam->getViewport());
            mDestRenderSystem->clearFrameBuffer (FBT_DEPTH | FBT_COLOUR | FBT_STENCIL, 
                                                c_cam->getViewport ()->getBackgroundColour (),//color
                                                1.0f,//depth
                                                0);//stencil
            mDestRenderSystem->_beginFrame();
            octree_assert (c_cam->getViewport()->getClearEveryFrame () == false);
        #endif //HWOCCLUSIONRTT

		mDestRenderSystem->_beginGeometryCount();

    }
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::disableHardwareOcclusionTests()
    {
        OctreeCHCCamera * const octreeCam = static_cast <OctreeCHCCamera *> (mCameraInProgress);
       	
		octree_assert (octreeCam->getViewport ());
        #ifdef HWOCCLUSIONRTT

                mDestRenderSystem->_endFrame();
                mDestRenderSystem->_setViewport(octreeCam->getViewport ());

        #else //HWOCCLUSIONRTT
                mDestRenderSystem->_endFrame();
		#endif //HWOCCLUSIONRTT

		if (mBeforeHWOCCamDetail != PM_SOLID)
		{
			mDestRenderSystem->clearFrameBuffer (FBT_DEPTH | FBT_COLOUR, 
												octreeCam->getViewport ()->getBackgroundColour (),//color
												1.0f,//depth
												0);//stencil
			// make sure we restore previous detail level.
			octreeCam->setPolygonMode (mBeforeHWOCCamDetail);
		}	        

		octree_assert (mCameraInProgress->getViewport()->getClearEveryFrame () == false);

		// Notify camera or vis faces
		octreeCam->_addCHCRenderedFaces(mDestRenderSystem->_getFaceCount()); 
	}
	//---------------------------------------------------------------------
	void OctreeCHCSceneManager::_renderScene(Camera* camera, Viewport* vp, bool includeOverlays)
	{
		// save previous cam clear mode.
		mBeforeHWOCCamClear = vp->getClearEveryFrame ();

		SceneManager::_renderScene(camera, vp,	includeOverlays);

		// restore previous cam clear mode.
		vp->setClearEveryFrame (mBeforeHWOCCamClear);
	}

	//---------------------------------------------------------------------
	bool OctreeCHCSceneManager::setRenderTemporalIdentification(OctreeCHCCamera * const cam, const bool onlyShadowCasters)
	{
		bool newFrame = true;

		const Ogre::Viewport * const v = cam->getViewport();
		const Ogre::RenderTarget * const t = v->getTarget();

		RenderTemporalIdentification *toAdd = 0;
		std::vector <RenderTemporalIdentification *>::const_iterator itUniqueRenderId = mRenderTemporalIdentifications.begin();
		for (;itUniqueRenderId != mRenderTemporalIdentifications.end(); ++itUniqueRenderId)
		{
			const RenderTemporalIdentification * const curr = *itUniqueRenderId;
			if (curr->mCamera == cam && 
				curr->mViewport == v && 
				curr->mTarget == t && 
				curr->mOnlyShadowCasters == onlyShadowCasters)
			{
				toAdd = *itUniqueRenderId;
				break;
			}
		}

		OctreeCHCCamera * const octreeCam = static_cast <OctreeCHCCamera *> (cam);
		if (toAdd == 0)
		{
			toAdd = new RenderTemporalIdentification();
			toAdd->mCamera = cam ;
			toAdd->mViewport = cam->getViewport();
			toAdd->mTarget = cam->getViewport()->getTarget();			
			toAdd->mOnlyShadowCasters = onlyShadowCasters;
			toAdd->id = mRenderTemporalIdentificationIdCount++;
			mRenderTemporalIdentifications.push_back(toAdd);
			toAdd->mLastFrameId = 0;
		}

		if (toAdd->mCullMode != cam->getOcclusionMode() || toAdd->mLastFrameId == 0)
		{
			toAdd->mCullMode = cam->getOcclusionMode();
			cam->setRenderId (toAdd->id);
			cam->registerInOcclusion();
			toAdd->mRegistered	= true;
			toAdd->mLastFrameId = Root::getSingleton().getCurrentFrameNumber ();
		}
		else
		{
			const unsigned int framesSceneId = Root::getSingleton().getCurrentFrameNumber ();
			if  (framesSceneId != toAdd->mLastFrameId)
			{
				toAdd->mLastFrameId = framesSceneId;
			}
			else
			{
				newFrame = false;
			}
			
			cam->setRenderId (toAdd->id);

			cam->setOcclusionModeDirect(toAdd->mCullMode);
		}
		return newFrame;
	}
	//---------------------------------------------------------------------
	void OctreeCHCSceneManager::removeRenderTemporalIdentification(OctreeCHCCamera * cam)
	{
		std::vector <RenderTemporalIdentification *>::iterator itUniqueRenderId = mRenderTemporalIdentifications.begin();
		while (itUniqueRenderId != mRenderTemporalIdentifications.end())
		{
			const RenderTemporalIdentification * const curr = *itUniqueRenderId;
			if (curr->mCamera == cam)
			{
				itUniqueRenderId = mRenderTemporalIdentifications.erase(itUniqueRenderId);
			}
			else
			{
				++itUniqueRenderId;
			}
		}
		// resize strictly making sure it's exact reduced size.
		std::vector <RenderTemporalIdentification *> resizedRenderTemporalIdentifications(mRenderTemporalIdentifications.begin(), mRenderTemporalIdentifications.end());
		mRenderTemporalIdentifications.swap(resizedRenderTemporalIdentifications );
	}
	//---------------------------------------------------------------------
    void OctreeCHCSceneManager::_findVisibleObjects(Ogre::Camera* cam, 
													Ogre::VisibleObjectsBoundsInfo * visibleBounds, 
													bool onlyShadowCasters)
	{
		OctreeCHCCamera * const octreeCam = static_cast <OctreeCHCCamera *> (cam);

		const bool newFrame = setRenderTemporalIdentification(octreeCam, onlyShadowCasters);

        RenderQueue *q = SceneManager::getRenderQueue();

		mCulling.mDisplayNode = mDisplayNodes;
		mCulling.mShowBoxes = mShowBoxes;

		// Make any pending updates to the calculated frustum planes
		const culling_modes cullMode = octreeCam->getOcclusionMode();
		if (cullMode == VIEW_FRUSTUM_COMPLEX || 
			cullMode == VIEW_FRUSTUM_RADAR ||
			cullMode == VIEW_FRUSTUM_MASK ||
			cullMode == VIEW_FRUSTUM_MASK_TEMPORAL)
		{
			// update special culling camera parameters (sphere frustum, cone frustum, radar info)
			octreeCam->updateCameraPrecalc();
		}

		mCulling.setOcclusionModeIsHWOcclusion(cullMode == CHC || cullMode == CHC_CONSERVATIVE || cullMode == STOP_AND_WAIT );
		
		mCamInProgressVisibles = octreeCam->getVisiblesList();
        octree_assert (mCamInProgressVisibles);

        mVisualDebugBoundingBoxes.clear();
        if (mCullCamera)
        {
            if (getCamera("CullCamera"))
                cam = getCamera("CullCamera");
        }

		
		mCulling.nextFrame (octreeCam, onlyShadowCasters, q);


		switch (cullMode)
		{
			case VIEW_FRUSTUM:
				mTreeManager->traverse (ViewFrustumCullingDepthTraversal (mCulling, visibleBounds));
				break;  
			case VIEW_FRUSTUM_MASK:
				mTreeManager->traverse (ViewFrustumCullingDepthTraversalPlaneMask (mCulling, visibleBounds));
				break;  
			case VIEW_FRUSTUM_MASK_TEMPORAL:
				mTreeManager->traverse (ViewFrustumCullingDepthTraversalPlaneMaskTemporal (mCulling, visibleBounds));
				break;  
			case VIEW_FRUSTUM_COMPLEX:
				mTreeManager->traverse (ViewFrustumCullingDepthTraversalComplex (mCulling, visibleBounds));
				break;  
			case VIEW_FRUSTUM_DIRECT:
				mTreeManager->traverse (ViewFrustumCullingDepthTraversalDirect (mCulling, visibleBounds));
				break;
			case VIEW_FRUSTUM_RADAR:
				mTreeManager->traverse (ViewFrustumCullingDepthTraversalRadar (mCulling, visibleBounds));
				break;
			case CHC:       
				{
					// ensure not Occlusion testing until we draw something
					mCulling.setFirstDrawPending(true);
					enableHardwareOcclusionTests();
					mCulling.CHCdepthTraversal (mTreeManager->getRoot(), visibleBounds);
					disableHardwareOcclusionTests();

					//q->clear();
				}
				break;
			case CHC_CONSERVATIVE:
				{
					// ensure not Occlusion testing until we draw something
					mCulling.setFirstDrawPending(true);
					enableHardwareOcclusionTests();
					mCulling.CHCdepthTraversalConservative (mTreeManager->getRoot(), visibleBounds);
					disableHardwareOcclusionTests();

					//q->clear();
				}
				break;
			case STOP_AND_WAIT:
				{     
					// ensure not Occlusion testing until we draw something
					mCulling.setFirstDrawPending(true);
					enableHardwareOcclusionTests();
					mCulling.initQueryPool ();
					mTreeManager->traverse (StopAndWaitTraversal (mCulling, visibleBounds));
					disableHardwareOcclusionTests(); 

					//q->clear();
				}
				break;

			default:
				octree_assert (0);
				break;
		}

        if (mCamInProgressVisibles && !mCamInProgressVisibles->empty())
        {
			//visibleBounds->reset();
            MovableObjectList::iterator movableIt = mCamInProgressVisibles->begin();
            MovableObjectList::iterator mitend = mCamInProgressVisibles->end(); 
            while (movableIt != mitend)
            {
				MovableObject * const m = (*movableIt);

				m->_updateRenderQueue (q);
				visibleBounds->merge(m->getWorldBoundingBox(true), 
									 m->getWorldBoundingSphere(true), 
									 cam);
                ++movableIt;
            }   
        }

		if (mCullDebugText) 
		{
			mDebugText = String("Drawn Nodes= ") + 
				StringConverter::toString(mCulling.mObjectCount) + ", Traversed=" +
				StringConverter::toString(mCulling.mTraversedNodesCount) + ", Culled=" +
				StringConverter::toString(mCulling.mFrustumCulledNodesCount) + ", Queried=" +
				StringConverter::toString(mCulling.mQueryCount) + ", Total Nodes=" + 
				StringConverter::toString(mTreeManager->getNumNodes()) + "  (" + 
				octreeCam->getOcclusionModeAsString() + ")" ;
				;
		}
		if (mCullDebug) 
			mTreeManager->traverse (TreeOverlayDebug (mCulling));
		
    }
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::addVisible(Ogre::MovableObject *mo)
    {   
        mCamInProgressVisibles->push_back (mo);
    }
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::directRenderSingleQueue(RenderQueue *queue)
    {  
        RenderQueue *oldQueue = mRenderQueue;
        mRenderQueue = queue;

        SceneManager::_renderVisibleObjects ();

        queue->clear ();
        mRenderQueue = oldQueue;
    }    
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::directRenderSingleObject(Renderable *r)
    {   
          const Camera * const c_cam = mCameraInProgress;
          mDestRenderSystem->_setWorldMatrix(Matrix4::IDENTITY);
//        mDestRenderSystem->_setViewMatrix(c_cam->getViewMatrix());
//        mDestRenderSystem->_setProjectionMatrix(c_cam->getProjectionMatrix());

		mDestRenderSystem->setCurrentPassIterationCount(1);
        _setPass(r->getMaterial()->getBestTechnique()->getPass(0));
        useRenderableViewProjMode (r);

        RenderOperation ro;
        r->getRenderOperation (ro);
        mDestRenderSystem->_render(ro);

    } 
	//---------------------------------------------------------------------
	OctreeCHCCamera* OctreeCHCSceneManager::getCurrentCurrentOptionCamera()
	{
		return mCurrentOptionCamera;
	}
	//---------------------------------------------------------------------
	void OctreeCHCSceneManager::setCurrentCurrentOptionCamera(OctreeCHCCamera* c)
	{
		mCurrentOptionCamera = c;
	}
    //---------------------------------------------------------------------
    bool OctreeCHCSceneManager::setOption(const Ogre::String & key, const void * val)
    {
        if (key == "Size")
        {
            octree_assert (mTreeManager);
            mTreeManager->resize(* static_cast < const Ogre::AxisAlignedBox * > (val));
            return true;
        }

        else if (key == "Depth")
        {
            octree_assert (mTreeManager);
			mTreeManager->setMaxDepth(* static_cast < const int * > (val));
			// copy the box since resize will delete mOctree and reference won't work
			Ogre::AxisAlignedBox box = mTreeManager->getBoundingBox();
            mTreeManager->resize(box);
            return true;
        }
        else if (key == "NextCullMode")
        {
            if (val)
            {
                CameraList::iterator it = mCameras.find (static_cast < const OctreeCHCCamera * > (val)->getName());
                if (it != mCameras.end())
				{
                    static_cast < OctreeCHCCamera * > (it->second)->setNextOcclusionMode();
                    return true;
                }
            }
            return false;
		}
		else if (key == "setNumberOfConservativeFrames")
		{
			mCulling.setNumberOfConservativeFrames(*(static_cast < const unsigned int * > (val)));
		}
		else if (key == "CurrentOptionCamera")
		{ 
			if (val)
			{
				CameraList::iterator it = mCameras.find (static_cast < const OctreeCHCCamera * > (val)->getName());
				if (it != mCameras.end())
				{
					mCurrentOptionCamera = static_cast < OctreeCHCCamera * > (it->second);
				}
				return true;
			}	
		}
		else if (key == "setCullingMode")
		{
			octree_assert (mCurrentOptionCamera);
			mCurrentOptionCamera->setOcclusionModeAsString(* static_cast < const Ogre::String * > (val));			
			return true;		
		}
        
            else if (key == "ShowOctree")
            {   
                mShowBoxes = * static_cast < const bool * > (val);
                return true;
            }
            else if (key == "CullDebug")
            {
                mCullDebug = * static_cast < const bool * > (val);
                return true;
			}
			else if (key == "CullDebugText")
			{
				mCullDebugText = * static_cast < const bool * > (val);
				return true;
			} 
            else if (key == "CullCamera")
            {
                mCullCamera = * static_cast < const bool * > (val);
                return true;
            }
            
        return SceneManager::setOption(key, val);


    }
    //---------------------------------------------------------------------
    bool OctreeCHCSceneManager::getOption(const Ogre::String & key, void *val)
    {
        if (key == "Size")
        {
			octree_assert(mTreeManager)
            Ogre::AxisAlignedBox * b = static_cast < Ogre::AxisAlignedBox * > (val);
            b->setExtents(mTreeManager->getBoundingBox ().getMinimum(), mTreeManager->getBoundingBox().getMaximum());
            return true;
        }

        else if (key == "Depth")
        {
            * static_cast < int * > (val) = mTreeManager->getMaxDepth();
            return true;
		}
		else if (key == "setNumberOfConservativeFrames")
		{
			* static_cast < unsigned int * > (val) = mCulling.getNumberOfConservativeFrames();
			return true;
		}  
		else if (key == "CurrentOptionCamera")
		{
			*static_cast < String * > (val) = mCurrentOptionCamera->getName(); 
			return true;	
		}
		else if (key == "getCullingMode")
		{
			octree_assert (mCurrentOptionCamera);
			*static_cast < String * > (val) = mCurrentOptionCamera->getOcclusionModeAsString();
			return true;		
		}
		
        
            else if (key == "ShowOctree")
            {

                * static_cast < bool * > (val) = mShowBoxes;
                return true;
            }
            else if (key == "CullCamera")
            {
                * static_cast < bool * > (val) = mCullCamera;
                return true;
            } 
            else if (key == "CullDebug")
            {
                * static_cast < bool * > (val) = mCullDebug;
                return true;
			}
			else if (key == "CullDebugText")
			{
				* static_cast < bool * > (val) = mCullDebugText;
				return true;
			}
			else if (key == "DebugText")
			{
				* static_cast < String * > (val) = mDebugText;
				return true;
			}
        
    
        return SceneManager::getOption(key, val);
    }
    //---------------------------------------------------------------------
    void OctreeCHCSceneManager::clearScene(void)
    {
        SceneManager::clearScene();

         // reset cameras
        //CameraList::iterator iCam = mCameras.begin();
        //for (; iCam != mCameras.end(); ++iCam)
        //{
        //    OctreeCHCCamera *cam = static_cast <OctreeCHCCamera *> (iCam->second);
        //    if (cam->isRegisteredInOcclusionSystem ())
        //    {
        //        mOctreeCHC->depthTraversal (UnregisterCameraDepthTraversal(cam));                   
        //        cam->setRegisteredInOcclusionSystem (false);
       //     } 
        //}           
		octree_assert(mTreeManager);
		AxisAlignedBox box  (mTreeManager->getBoundingBox());
        init(box, mTreeManager->getMaxDepth());

    }
    //---------------------------------------------------------------------
    Ogre::AxisAlignedBoxSceneQuery*
    OctreeCHCSceneManager::createAABBQuery(const Ogre::AxisAlignedBox& box, unsigned long mask)
    {
        OctreeCHCAxisAlignedBoxSceneQuery* q = new OctreeCHCAxisAlignedBoxSceneQuery(this);
        q->setBox(box);
        q->setQueryMask(mask);
        return q;
    }
    //---------------------------------------------------------------------
    SphereSceneQuery*
    OctreeCHCSceneManager::createSphereQuery(const Sphere& sphere, unsigned long mask)
    {
        OctreeCHCSphereSceneQuery* q = new OctreeCHCSphereSceneQuery(this);
        q->setSphere(sphere);
        q->setQueryMask(mask);
        return q;
    }
    //---------------------------------------------------------------------
    PlaneBoundedVolumeListSceneQuery*
    OctreeCHCSceneManager::createPlaneBoundedVolumeQuery(const PlaneBoundedVolumeList& volumes,
            unsigned long mask)
    {
        OctreeCHCPlaneBoundedVolumeListSceneQuery* q = new OctreeCHCPlaneBoundedVolumeListSceneQuery(this);
        q->setVolumes(volumes);
        q->setQueryMask(mask);
        return q;
    }

    //---------------------------------------------------------------------
    RaySceneQuery*
    OctreeCHCSceneManager::createRayQuery(const Ray& ray, unsigned long mask)
    {
        OctreeCHCRaySceneQuery* q = new OctreeCHCRaySceneQuery(this);
        q->setRay(ray);
        q->setQueryMask(mask);
        return q;
    }
    //---------------------------------------------------------------------
    IntersectionSceneQuery*
    OctreeCHCSceneManager::createIntersectionQuery(unsigned long mask)
   {

    // OctreeCHC implementation performs WORSE for < 500 objects
    // TODO: optimize it so it's better in all cases
    //OctreeCHCIntersectionSceneQuery* q = new OctreeCHCIntersectionSceneQuery(this);
    DefaultIntersectionSceneQuery* q;
    //if (mOctreeCHC->numNodes () > 500)
        q = new OctreeCHCIntersectionSceneQuery(this);
    //else
   //     q = new DefaultIntersectionSceneQuery(this);
    q->setQueryMask(mask);
    return q;
   }
    //---------------------------------------------------------------------
	void OctreeCHCSceneManager::resize(const Ogre::AxisAlignedBox& box)
	{
		mTreeManager->resize(box);
	};    
    //---------------------------------------------------------------------
	void OctreeCHCSceneManager::resize(const Ogre::AxisAlignedBox &box, const int depth)
	{
		mTreeManager->resize(box, depth);
	};
    //---------------------------------------------------------------------
	int OctreeCHCSceneManager::getMaxDepth() const 
	{ 
		return mTreeManager->getMaxDepth(); 
	}
    //---------------------------------------------------------------------
	void OctreeCHCSceneManager::setMaxDepth(int val) 
	{ 
		mTreeManager->setMaxDepth(val); 
	}
    //---------------------------------------------------------------------
	const Ogre::AxisAlignedBox &OctreeCHCSceneManager::getBoundingBox()const 
	{
		return mTreeManager->getBoundingBox();
	};

}
